import datetime
import time
import iso8601

import psycopg2
import psycopg2.pool
import psycopg2.extras

from edge import eventlog
from edge.com import db
from edge.types import JnodeError
from edge_bbox.types import QuadI

from edge.webserver.systest import api
from edge.webserver.systest.test import Test

from .stats import StoreCounters
from .types import StoreArgs
from .._models import Fact


class StoreError(Exception):
    pass


def _isotime_to_ts(iso_time):
    """
    :type iso_time: str
    :rtype: int
    """
    dt = iso8601.parse_date(iso_time)
    epoch_dt = datetime.datetime.fromtimestamp(0, dt.tzinfo)
    ts_sec = (dt - epoch_dt).total_seconds()
    return int(ts_sec * 1e6)


class _Limits:
    def __init__(self, ts, origin, quad):
        """
        :type ts: long
        :type origin: str
        :type quad: QuadI | None
        """
        self.ts = ts
        self.origin = origin
        self.quad = quad


class _MinFact:
    def __init__(self, ts, origin, scene_ref, plate_text, plate_quad, speed_dir, pos_fl):
        """
        :type ts: int
        :type origin: str
        :type scene_ref: str
        :type plate_text: str
        :type plate_quad: QuadI
        :type speed_dir: (int, str) | None
        :type pos_fl: (float, float) | None
        """
        self.ts = ts
        self.origin = origin
        self.scene_ref = scene_ref

        self.plate_text = plate_text
        self.plate_quad = plate_quad
        self.speed_dir = speed_dir
        self.pos_fl = pos_fl

    @staticmethod
    def unpack(jnode):
        """
        :param jnode: args.json#/
        :type jnode: dict

        :rtype: _MinFact
        :raise JnodeError:
        """
        if jnode['type'] != 'transit':
            eventlog.log_i('fail: param #/type has invalid value <{}>'.format(jnode['type']))
            raise JnodeError

        ts = _isotime_to_ts(jnode['time'])
        origin = jnode['origin']['uuid']

        m_jnodes = jnode['measurements']
        m_jnode = m_jnodes[0]
        b_jnode = m_jnode['waypoints']['best']
        v_jnode = b_jnode['vehicle']
        p_jnode = v_jnode['license-plates'][0]

        # unpack plate
        v_plate_text = p_jnode['text']['ucode']
        v_plate_quad = QuadI.unpack(p_jnode['bbox-4'])

        # unpack speed, dir
        v_speed_dir = tuple([
            v_jnode.get('spot-speed', {}).get('radar', {}).get('measured'),
            v_jnode.get('dir')
        ])

        if None in v_speed_dir:
            v_speed_dir = None

        # unpack position
        v_pos_jnode = v_jnode.get('position', {})
        forward_sm = v_pos_jnode.get('forward')
        left_sm = v_pos_jnode.get('left')
        v_pos_fl = tuple([
            forward_sm / 100.0 if forward_sm else None,
            left_sm / 100.0 if left_sm else None
        ])

        # if ra-measured speed is None - position values are not accurate
        if None in v_pos_fl or v_speed_dir is None:
            v_pos_fl = None

        # unpack scene
        scene_ref = b_jnode.get('images', {}).get('scene', {}).get('name')
        if not scene_ref:
            eventlog.log_e_param('#/measurements/0/waypoints/best/images/scene/name')
            raise JnodeError

        return _MinFact(ts, origin, scene_ref, v_plate_text, v_plate_quad, v_speed_dir, v_pos_fl)


class IStoreService:
    def __init__(self):
        pass

    def setup(self, args):
        """
        :type args: dict
        :raise StoreError:
        """
        pass

    def save(self, fact):
        """
        :type fact: Fact

        :return: (platear-id, media-id)
        :rtype: (int, int) | None

        :raise StoreError:
        """
        raise NotImplementedError

    def destroy(self):
        pass


class _FactStore(IStoreService):
    def __init__(self, counters, args):
        """
        :type counters: StoreCounters
        :type args: db.DbArgs
        """
        IStoreService.__init__(self)

        try:
            self._link_pool = psycopg2.pool.SimpleConnectionPool(
                minconn=1,
                maxconn=1,
                user=args.db_user,
                password=args.db_pass,
                database=args.db_name,
                host=args.db_host,
                port=args.db_port)

        except psycopg2.Error as x:
            eventlog.log_e('fail: create db-link pool <{}>'.format(x.message))
            raise StoreError

        self._counters = counters

    def destroy(self):
        self._link_pool.closeall()

    @staticmethod
    def _save(db_link, fact):
        """
        :type fact: Fact

        :rtype: (int, int)
        :return: Tuple (platear-id, media-id)

        :raise StoreError:
        """
        query_jnode = {
            'event': fact.jnode,
            'dist': {
                'list': fact.arch_files,
                'selected': fact.feed_files,
                'store': fact.args_store
            }
        }

        with db_link.cursor() as db_cursor:
            db_cursor.execute(
                "SELECT errcode,errmsg,result FROM api_event_add(%s)",
                [db.DbJsonArg(query_jnode)])
            errcode, errmsg, reply_jnode = db_cursor.fetchone()

            if errcode == 0:
                platear_id = reply_jnode['id']
                scroll_id = reply_jnode['scrollid']
                return platear_id, scroll_id

            else:
                eventlog.log_e('fail: api_event_add (code=<{}>, text=<{}>)'.format(errcode, errmsg))
                raise StoreError

    def save(self, fact):
        """
        :type fact: Fact

        :return: (platear-id, media-id)
        :rtype: (int, int) | None

        :raises StoreError:
        """
        try:
            db_link = self._link_pool.getconn()
            self._counters.inc_link_present()
            try:
                with db_link:
                    platear_id, scroll_id = self._save(db_link, fact)
                    self._counters.last_event_time()
                    return platear_id, scroll_id

            finally:
                self._link_pool.putconn(db_link)

        except psycopg2.pool.PoolError as x:
            eventlog.log_e('fail: db-link <{}>'.format(x.message))
            self._counters.inc_link_absent()
            raise StoreError


class _TestStore(IStoreService):
    def __init__(self, args):
        """
        :type args: db.DbArgs
        """
        IStoreService.__init__(self)

        self._testpack = api.init(args)
        self._ref_to_limits_map = dict()

    def destroy(self):
        api.fini(self._testpack)

    def setup(self, args):
        """
        :type args: dict
        :raise StoreError:
        """
        call_ts = int(time.time())  # sec

        origin = args['origin']
        listen_tout = args['tout']  # sec
        session_ref = args['session-ref']
        test_tef = tuple(args['test-ref'])
        bbox_jnode = args.get('bbox4')

        if listen_tout == 0:
            self._ref_to_limits_map = dict()
        else:
            self._ref_to_limits_map[(session_ref, test_tef)] = \
                _Limits(1000000 * (call_ts + listen_tout), origin, QuadI.unpack(bbox_jnode) if bbox_jnode else None)

    def save(self, fact):
        """
        :type fact: Fact

        :return: (platear-id, media-id)
        :rtype: (int, int) | None

        :raise StoreError:
        """
        # unpack
        try:
            min_fact = _MinFact.unpack(fact.jnode)

        except (KeyError, ValueError) as x:
            eventlog.log_e('invalid args.json <{}>'.format(x.message))
            return
        except JnodeError:
            return

        # inspect
        tout_refs = []
        save_refs = []
        for ref, limits in self._ref_to_limits_map.iteritems():
            if min_fact.ts > limits.ts:
                tout_refs.append(ref)

            elif min_fact.origin == limits.origin:
                if limits.quad is None or min_fact.plate_quad.center in limits.quad:
                    save_refs.append(ref)

        # delete
        for ref in tout_refs:
            del self._ref_to_limits_map[ref]

        # update
        for ref in save_refs:
            session_ref, test_ref = ref
            session = self._testpack.get_session(session_ref)
            test = session.get_test(test_ref)  # type: Test

            test_type = test_ref[0]
            if test_type == 'spot-speed':
                if min_fact.speed_dir:
                    data_jnode = {
                        'plate': min_fact.plate_text,
                        'images': {
                            'scene': {
                                'ref': min_fact.scene_ref
                            }
                        },
                        'speed': {
                            'actual': min_fact.speed_dir[0]
                        },
                        'dir': {
                            'actual': min_fact.speed_dir[1]
                        }
                    }

                    test.add_step({'data': data_jnode})

            elif test_type == 'position':
                if min_fact.pos_fl:
                    data_jnode = {
                        'plate': min_fact.plate_text,
                        'images': {
                            'scene': {
                                'ref': min_fact.scene_ref
                            }
                        },
                        'forward': {
                            'actual': min_fact.pos_fl[0]
                        },
                        'left': {
                            'actual': min_fact.pos_fl[1]
                        }
                    }

                    test.add_step({'data': data_jnode})

            else:
                raise NotImplementedError


def init_service(counters, args):
    """
    :type counters: StoreCounters
    :type args: StoreArgs

    :rtype: IStoreService
    :raise StoreError:
    """
    if args.systest:
        service = _TestStore(args.db)
    else:
        service = _FactStore(counters, args.db)

    return service


def fini_service(service):
    """
    :type service: IStoreService | _FactStore | _TestStore
    """
    service.destroy()
