import os

from edge import eventlog
from edge.types import JnodeError
from edge.service.types import AboutServiceArgs
from edge.service.jnode.types import JsonGpsdServiceArgs, JsonTzServiceArgs, TzServiceArgs, GpsdServiceArgs
from edge.image.jnode.types import JsonImFooterArgs, ImFooterArgs

from .tl import JsonTlServiceArgs

class StatsServiceArgs:
    def __init__(self, stat_filepath, stat_tout_sec):
        """
        :type stat_filepath: str
        :type stat_tout_sec: int
        """
        self.stat_filepath = stat_filepath
        self.stat_tout_sec = stat_tout_sec

    @staticmethod
    def deserialize(jnode):
        """
        :param jnode: main.conf#/services/monitoring
        :type jnode: dict

        :rtype: StatsServiceArgs
        :raise JnodeError:
        """
        stat_path = jnode.get('path')
        if not (type(stat_path) in [unicode, str] and len(stat_path) < 64 and os.path.isabs(stat_path)):
            eventlog.log_e_param('#/path')
            raise JnodeError

        stat_tout_sec = jnode.get('tout')
        if not (int == type(stat_tout_sec) and 0 <= stat_tout_sec <= 255):
            eventlog.log_e_param('#/tout')
            raise JnodeError

        return StatsServiceArgs(stat_path, stat_tout_sec)


class ReactArgs:
    def __init__(self, endpoint):
        """
        :type endpoint: str
        """
        self.endpoints = [endpoint]

    @staticmethod
    def unpack(jnode):
        """
        :type jnode: dict
        :rtype: ReactArgs
        """
        return ReactArgs(jnode['Url'])


class ImLoaderArgs:
    def __init__(self, dh, image_wh):
        """
        :type dh: int
        :type image_wh: (int, int)
        """
        self.dh = dh
        self.image_wh = image_wh


class ImStoreArgs:
    def __init__(self, im_format, im_q):
        """
        :type im_format: str
        :type im_q: ImQueueArgs

        """
        self.codec = im_format
        self.image_queue_args = im_q

    @staticmethod
    def unpack(jnode):
        """
        :param jnode: main.conf#/image-store
        :type jnode: dict

        :rtype: ImStoreArgs
        :raise JnodeError:
        """
        im_format = jnode.get('format')
        if im_format not in ['JPEG', 'RGBA']:
            eventlog.log_e_param('#/image-store')
            raise JnodeError

        im_max_count = jnode.get('max-file-count', 4)
        if not (int == type(im_max_count) and 4 <= im_max_count <= 8):
            eventlog.log_e_param('#/max-file-count')
            raise JnodeError

        endpoint = jnode.get('endpoint')
        if not (type(endpoint) in [unicode, str] and len(endpoint) < 64 and os.path.isabs(endpoint)):
            eventlog.log_e_param('#/endpoint')
            raise JnodeError

        im_q = ImQueueArgs(endpoint, im_max_count - 2, im_max_count)
        return ImStoreArgs(im_format, im_q)


class ImQueueArgs:
    def __init__(self, endpoint, min_count, max_count):
        """
        :type endpoint: str
        :type min_count: int
        :type max_count: int
        """
        self.endpoint = endpoint
        self.min_file_count = min_count
        self.max_file_count = max_count


class JsonQueueArgs:
    def __init__(self, endpoint, min_count, max_count):
        """
        :type endpoint: str
        :type min_count: int
        :type max_count: int
        """
        self.endpoint = endpoint
        self.min_file_count = min_count
        self.max_file_count = max_count


class EventArgs:
    def __init__(self, q_args, log_level='info'):
        """
        :type q_args: list[JsonQueueArgs]
        :type log_level: str
        """
        self.json_queue_args_list = q_args
        self.log_level = log_level

    @staticmethod
    def unpack(jnode, log_level):
        """
        :param jnode: main.conf#/event-store
        :type jnode: dict
        :type log_level: str

        :rtype: EventArgs
        :raise JnodeError:
        """
        max_file_count = jnode.get('max-file-count', 4)
        if not (int == type(max_file_count) and 4 <= max_file_count <= 8):
            eventlog.log_e_param('#/max-file-count')
            raise JnodeError

        q_args = map(
            lambda endpoint: JsonQueueArgs(endpoint, max_file_count - 5, max_file_count),
            jnode.get('endpoints') or list())

        return EventArgs(q_args, log_level)


class LrAnnoArgs:
    _MODE_NEVER = 'never'
    _MODE_ALWAYS = 'always'

    def __init__(self,
                 dir_mode='never',
                 lane_mode='never',
                 text_mode='never',
                 trace_mode='never',
                 margin_mode='never',
                 center_mode='never'):
        """
        :type dir_mode: str
        :type lane_mode: str
        :type text_mode: str
        :type trace_mode: str
        :type margin_mode: str
        :type center_mode: str
        """
        self.show_dir = dir_mode == 'always'
        self.show_lane = lane_mode == 'always'
        self.show_text = text_mode == 'always'
        self.show_trace = trace_mode == 'always'
        self.show_margin = margin_mode == 'always'
        self.show_center = center_mode == 'always'

    @classmethod
    def unpack(cls, jnode):
        """
        :param jnode: main.conf#/Renderer/profile/vehicle/lr
        :type jnode: dict

        :rtype: LrAnnoArgs
        :raise JnodeError:
        """
        modes = [cls._MODE_NEVER, cls._MODE_ALWAYS]

        # unpack #/plate
        plate_mode = jnode.get('plate', cls._MODE_NEVER)
        if plate_mode not in modes:
            eventlog.log_e_param('#/plate')
            raise JnodeError

        # unpack #/trace
        trace_mode = jnode.get('trace', cls._MODE_NEVER)
        if trace_mode not in modes:
            eventlog.log_e_param('#/trace')
            raise JnodeError

        # unpack #/body
        margin_mode = jnode.get('body', cls._MODE_NEVER)
        if margin_mode not in modes:
            eventlog.log_e_param('#/body')
            raise JnodeError

        # unpack #/dir
        dir_mode = jnode.get('dir', cls._MODE_NEVER)
        if dir_mode not in modes:
            eventlog.log_e_param('#/dir')
            raise JnodeError

        # unpack #/lane
        lane_mode = jnode.get('lane', cls._MODE_NEVER)
        if lane_mode not in modes:
            eventlog.log_e_param('#/lane')
            raise JnodeError

        # unpack #/center
        center_mode = jnode.get('center', cls._MODE_NEVER)
        if margin_mode not in modes:
            eventlog.log_e_param('#/center')
            raise JnodeError

        return LrAnnoArgs(dir_mode, lane_mode, plate_mode, trace_mode, margin_mode, center_mode)


class RaAnnoArgs:
    _MODE_IF_PLATE = 'if-plate'
    _MODE_ALWAYS = 'always'
    _MODE_NEVER = 'never'

    def __init__(self,
                 yellow_speed,
                 red_speed,
                 len_mode='never',
                 dir_mode='never',
                 pos_mode='never',
                 speed_mode='never'):
        """
        :type yellow_speed: int
        :type red_speed: int
        :type speed_mode: str
        :type len_mode: str
        :type dir_mode: str
        :type pos_mode: str
        """
        self._speed_mode = speed_mode
        self._len_mode = len_mode
        self._dir_mode = dir_mode
        self._pos_mode = pos_mode

        self.yellow_speed = yellow_speed
        self.red_speed = red_speed

    @classmethod
    def _show(cls, mode, lr_ra_match):
        """
        :type lr_ra_match: bool
        :rtype: bool
        """
        if mode == cls._MODE_ALWAYS:
            return True

        elif mode == cls._MODE_IF_PLATE:
            return lr_ra_match

        else:
            return False

    def show_len(self, lr_ra_match=False):
        """
        :type lr_ra_match: bool
        :rtype: bool
        """
        return self._show(self._len_mode, lr_ra_match)

    def show_dir(self, lr_ra_match=False):
        """
        :type lr_ra_match: bool
        :rtype: bool
        """
        return self._show(self._dir_mode, lr_ra_match)

    def show_pos(self, lr_ra_match=False):
        """
        :type lr_ra_match: bool
        :rtype: bool
        """
        return self._show(self._pos_mode, lr_ra_match)

    def show_speed(self, lr_ra_match=False):
        """
        :type lr_ra_match: bool
        :rtype: bool
        """
        return self._show(self._speed_mode, lr_ra_match)

    @classmethod
    def unpack(cls, jnode, yellow_speed, red_speed):
        """
        :param jnode: main.conf#/Renderer/profile/vehicle/ra
        :type jnode: dict
        :type yellow_speed: int
        :type red_speed: int

        :rtype: LrAnnoArgs
        :raise JnodeError:
        """
        modes = [cls._MODE_NEVER, cls._MODE_IF_PLATE, cls._MODE_ALWAYS]

        # unpack #/pos
        pos_mode = jnode.get('pos', cls._MODE_NEVER)
        if pos_mode not in modes:
            eventlog.log_e_param('#/pos')
            raise JnodeError

        # unpack #/speed
        speed_mode = jnode.get('speed', cls._MODE_NEVER)
        if speed_mode not in modes:
            eventlog.log_e_param('#/speed')
            raise JnodeError

        # unpack #/plate
        dir_mode = jnode.get('dir', cls._MODE_NEVER)
        if dir_mode not in modes:
            eventlog.log_e_param('#/dir')
            raise JnodeError

        # unpack #/len
        len_mode = jnode.get('len', cls._MODE_NEVER)
        if len_mode not in modes:
            eventlog.log_e_param('#/len')
            raise JnodeError

        return RaAnnoArgs(yellow_speed, red_speed, len_mode, dir_mode, pos_mode, speed_mode)


class RendererArgs:
    def __init__(self, lr_anno, ra_anno):
        """
        :type lr_anno: LrAnnoArgs
        :type ra_anno: RaAnnoArgs
        """
        self.lr_anno = lr_anno
        self.ra_anno = ra_anno

    @classmethod
    def unpack(cls, jnode, yellow_speed, red_speed):
        """
        :param jnode: main.conf#/Renderer/profile/vehicle
        :type jnode: dict
        :type yellow_speed: int
        :type red_speed: int
        :rtype: RendererArgs
        """
        lr_jnode = jnode.get('lr')
        lr_anno = LrAnnoArgs.unpack(lr_jnode) if lr_jnode else LrAnnoArgs()

        ra_jnode = jnode.get('ra')
        if ra_jnode:
            ra_anno = RaAnnoArgs.unpack(ra_jnode, yellow_speed, red_speed)
        else:
            ra_anno = RaAnnoArgs(yellow_speed, red_speed)

        return RendererArgs(lr_anno, ra_anno)


class LaneDetectorArgs:
    def __init__(self, lane_to_mask_map, image_w, image_h):
        """
        :type lane_to_mask_map: list[(int, list[(int, int)])] | list[(str, list[(int, int)])]
        :type image_w: int
        :type image_h: int
        """
        self.id_to_lane_list = lane_to_mask_map
        self.img_width = image_w
        self.img_height = image_h

    @staticmethod
    def unpack(jnode):
        """
        :param jnode: main.conf#/road-mask
        :type jnode: dict
        :rtype: LaneDetectorArgs
        """
        lanes_jnode = jnode.get('lanes', [])
        lane_to_mask_list = map(lambda lane_jnode: (lane_jnode['name'], lane_jnode['mask']), lanes_jnode)
        img_w, img_h = jnode.get('scene-wh', [0, 0])

        return LaneDetectorArgs(lane_to_mask_list, img_w, img_h)


class KernelArgs:
    def __init__(self, react, event, image_store, renderer, lane_detector_args, lr_store_tout, tz, gpsd,
                 footer, monitoring, im_loader, about, tl_service_args):
        """
        :type react: ReactArgs
        :type event: EventArgs
        :type image_store: ImStoreArgs
        :type renderer: RendererArgs
        :type lane_detector_args: LaneDetectorArgs
        :type lr_store_tout: int
        :type tz: TzServiceArgs
        :type gpsd: GpsdServiceArgs
        :type footer: ImFooterArgs
        :type monitoring: StatsServiceArgs
        :type im_loader: ImLoaderArgs
        :type about: AboutServiceArgs
        :type tl_service_args: JsonTlServiceArgs | None
        """
        self.lane_detector_args = lane_detector_args
        self.react = react
        self.scene_event = event
        self.image_store = image_store
        self.renderer = renderer
        self.lr_store_tout = lr_store_tout
        self.tz = tz
        self.gpsd = gpsd
        self.footer = footer
        self.monitoring = monitoring
        self.im_loader = im_loader
        self.about = about
        self.tl = tl_service_args

    @staticmethod
    def unpack(jnode, log_level):
        """
        :param jnode: main.conf#/
        :type jnode: dict
        :type log_level: str

        :rtype: KernelArgs
        :raise JnodeArgs:
        """
        lane_detector = LaneDetectorArgs.unpack(jnode.get('road-mask', {}))

        rn_jnode = jnode['Renderer']
        profile_jnode = rn_jnode['profile']
        footer = JsonImFooterArgs.deserialize(profile_jnode['footer'])

        react = ReactArgs.unpack(jnode['Source'])
        event = EventArgs.unpack(jnode['scene-event'], log_level)

        im_loader = ImLoaderArgs(footer.height, (profile_jnode['width'], 65535))
        im_store = ImStoreArgs.unpack(jnode['image-store'])

        vehicle_jnode = profile_jnode.get('annotation', {}).get('vehicle', {})

        tl_service_args = JsonTlServiceArgs.deserialize(jnode.get('tl-args'))

        yellow = rn_jnode.get('YellowSpeedThreshold', 65535)
        red = rn_jnode.get('RedSpeedThreshold', 65535)
        renderer = RendererArgs.unpack(vehicle_jnode, yellow, red)

        # define LrStoreTout
        def_lr_store_tout = 1
        lr_store_tout = rn_jnode.get('LrStoreTout', def_lr_store_tout)
        if not (1 <= lr_store_tout <= 60):

            eventlog.log_w('The param LrStoreTout has invalid value %d. The default value %d is used.'
                           % (lr_store_tout, def_lr_store_tout))
            lr_store_tout = def_lr_store_tout

        # deserialize services_jnode
        services_jnode = jnode['services']

        conf_path = services_jnode['origin']
        about = AboutServiceArgs(conf_path)

        # datetime service
        conf_path = services_jnode.get('datetime')
        if not conf_path:
            eventlog.log_e_param('#/services/datetime')
            raise JnodeError

        tz = JsonTzServiceArgs.create(conf_path)

        # geolocation service
        conf_path = services_jnode.get('geolocation')
        if not conf_path:
            eventlog.log_e_param('#/services/geolocation')
            raise JnodeError

        gpsd = JsonGpsdServiceArgs.create(conf_path)

        # stat service
        conf_jnode = services_jnode['monitoring']
        stats = StatsServiceArgs.deserialize(conf_jnode)

        return KernelArgs(react, event, im_store, renderer, lane_detector, lr_store_tout, tz, gpsd, footer,
                          stats, im_loader, about, tl_service_args)
