from twisted.internet import inotify
from twisted.python.filepath import FilePath

from edge import eventlog
from edge.com import fs
from edge.types import JnodeError

from edge.model.types import TimeSpan


class TlServiceArgs:
    def __init__(self, endpoint, tl_ident):
        """
        :type endpoint: str
        :type tl_ident: int
        """
        self.endpoint = endpoint
        self.tl_ident = tl_ident


class JsonTlServiceArgs(TlServiceArgs):
    def __init__(self, *args, **kwargs):
        TlServiceArgs.__init__(self, *args, **kwargs)

    @staticmethod
    def deserialize(jnode):
        """
        :type jnode: dict | None
        :rtype: JsonTlServiceArgs | None
        :raise JnodeError:
        """
        eventlog.log_i('jnode {}'.format(jnode))
        if jnode is None:
            return None

        if len(jnode) == 2:
            endpoint, tl_ident = jnode
            return JsonTlServiceArgs(endpoint, tl_ident)

        else:
            eventlog.log_e("fail: tl args invalid")
            return None


class ITlService:
    def __init__(self):
        pass

    def check(self, min_ts, max_ts):
        """
        :type min_ts: long
        :type max_ts: long
        :rtype: list[TimeSpan]
        """
        raise NotImplementedError

    def get_times(self, ts):
        """
        :param ts: value in usec
        :type ts: long

        :return: (green-amber-time, red-time, red-amber-time). Values in usec
        :rtype: (long, long, long)
        """
        raise NotImplementedError

    def get_times2(self, ts):
        """
        :param ts: value in usec
        :type ts: long

        :return: (green-time, red-time, amber-time). Values in usec
        :rtype: (long, long, long)
        """
        raise NotImplementedError

    def destroy(self):
        pass


def init_service(twisted_reactor, args):
    """
    :type twisted_reactor: twisted.internet.interfaces.IReactorUNIX
    :type args: JsonTlServiceArgs | None
    :rtype: ITlService
    """
    if args is None:
        rl = _StubService()
    else:
        rl = _TlService(twisted_reactor, args)
    return rl


def fini_service(service):
    """
    :type service: ITlService | _TlService
    """
    service.destroy()


class _StubService(ITlService):
    def __init__(self):
        ITlService.__init__(self)

    def check(self, min_ts, max_ts):
        return []

    def get_times(self, ts):
        return 0, 0, 0

    def get_times2(self, ts):
        return 0, 0, 0


def _print_ti(string, ti_l):
    """
    :type ti_l:list[TimeSpan]
    """
    print string
    for ti in ti_l:  # type: TimeSpan
        print ti.min_ts, '-', ti.max_ts


class _TrafficLight:
    def __init__(self, ch_id, red, amber, green=list()):
        """
        :type ch_id: int
        :type red: list[TimeSpan]
        :type amber: list[TimeSpan]
        :type green: list[TimeSpan]
        """
        self.ch_id = ch_id
        self.red = red
        self.amber = amber
        self.green = green

        self.prohibiting = self.red

    def get_red2(self, ts):
        """
        :type ts: long
        :rtype: long
        """
        red_spans = filter(lambda time_span: ts in time_span, self.red)

        if not len(red_spans):
            # current collor is green
            return 0

        # red started before ts
        red_spans = filter(lambda span: span.min_ts < ts, red_spans)  # type: list[TimeSpan]
        if not len(red_spans):
            return 0

        # check "just red" instead of "any red"
        # just_red = any_red - any_amber
        any_red_time = ts - red_spans[0].min_ts
        return any_red_time

    def get_green2(self, ts):
        """
        :type ts: long
        :rtype: long
        """
        green_spans = filter(lambda time_span: ts in time_span, self.green)

        if not len(green_spans):
            # current collor is green
            return 0

        # green started before ts
        green_spans = filter(lambda span: span.min_ts < ts, green_spans)  # type: list[TimeSpan]
        if not len(green_spans):
            return 0

        any_green_time = ts - green_spans[0].min_ts
        return any_green_time

    def get_amber2(self, ts):
        """
        :type ts: long
        :rtype: long
        """
        amber_spans = filter(lambda time_span: ts in time_span, self.amber)

        if not len(amber_spans):
            # current collor is green
            return 0

        # green started before ts
        amber_spans = filter(lambda span: span.min_ts < ts, amber_spans)  # type: list[TimeSpan]
        if not len(amber_spans):
            return 0

        any_green_time = ts - amber_spans[0].min_ts
        return any_green_time


class TlLog:
    def __init__(self, time, channels):
        """
        :type time: TimeSpan
        :type channels: list[_TrafficLight]
        """
        self.channels = channels
        self.time = time

    @staticmethod
    def deserialize(tl_ident, jnode):
        """
        :type tl_ident: int
        :type jnode: dict
        :rtype: JsonTlArgs

        :raise JnodeError:
        """
        ts_jnode = jnode.get('ts')

        if type(ts_jnode) != list or len(ts_jnode) != 2:
            eventlog.log_e_param('#/ts')
            raise JnodeError

        time_interval = TimeSpan(*ts_jnode)

        traffic_lights_jnode = jnode.get('tl')
        if type(traffic_lights_jnode) != list:
            eventlog.log_e_param('#/tl')

        traffic_lights = []
        for tl_jnode in traffic_lights_jnode:
            if tl_ident != tl_jnode['id']:
                eventlog.log_i('tl_ident != tl_jnode[id]')
                continue

            red = tl_jnode.get('red', [])
            amber = tl_jnode.get('amber', [])
            green = tl_jnode.get('green', [])

            # eventlog.log_i('{} {} {}'.format(len(red), len(amber), len(green)))

            red_intervals = map(TimeSpan.create, red)
            amber_intervals = map(TimeSpan.create, amber)
            green_intervals = map(TimeSpan.create, green)

            traffic_light = _TrafficLight(tl_ident, red_intervals, amber_intervals, green_intervals)
            traffic_lights.append(traffic_light)

        if not traffic_lights:
            eventlog.log_d('fail: duplicate idents found')
            raise JnodeError

        return TlLog(time_interval, traffic_lights)


class _TlService(ITlService):
    """
    :type _tl_log: TlLog | None
    """
    def __init__(self, twisted_reactor, args):
        """
        :type twisted_reactor: twisted.internet.interfaces.IReactorUNIX
        :type args: JsonTlServiceArgs
        """
        ITlService.__init__(self)

        self._twisted_reactor = twisted_reactor

        self._tl_log = None
        self._tl_ident = args.tl_ident

        self._endpoint = args.endpoint
        self._inotify = inotify.INotify()
        self._inotify.startReading()
        self._reset_inotify_watch()

    def destroy(self):
        try:
            self._inotify.ignore(FilePath(self._endpoint))
        except KeyError:
            pass

    def _reset_inotify_watch(self):
        try:
            self._inotify.ignore(FilePath(self._endpoint))
        except KeyError:
            pass

        fs.silent_mkdir_r(self._endpoint)

        try:
            self._inotify.watch(
                FilePath(self._endpoint),
                mask=inotify.IN_MOVED_TO | inotify.IN_CLOSE_WRITE | inotify.IN_DELETE_SELF | inotify.IN_MOVE_SELF,
                callbacks=[self._handle_react])
        except inotify.INotifyError as x:
            eventlog.log_d('fail: reset inotify watch (%s)' % x)

    # noinspection PyUnusedLocal
    def _handle_react(self, ignored, filepath, mask):
        if (mask & inotify.IN_DELETE_SELF) or (mask & inotify.IN_MOVE_SELF):
            self._twisted_reactor.callLater(0, self._reset_inotify_watch)
            eventlog.log_w('endpoint was changed to %s' % self._endpoint)
            return

        try:
            react_jnode = fs.load_json3(filepath.path)
            if react_jnode:
                self._tl_log = TlLog.deserialize(self._tl_ident, react_jnode)

        except JnodeError as x:
            return

        finally:
            fs.silent_rm_r(filepath.path)

    def get_times2(self, ts):
        """
        :param ts: value in usec
        :type ts: long

        :return: (green-time, red-time, amber-time). Values in usec
        :rtype: (long, long, long)
        """
        if self._tl_log is None:
            eventlog.log_d('fail: tl log is absent')
            return 0, 0, 0

        traffic_lights = filter(lambda ch: ch.ch_id == self._tl_ident, self._tl_log.channels)
        assert 1 == len(traffic_lights)

        tl = traffic_lights[0]  # type: _TrafficLight
        # eventlog.log_d('ts {} gr {} red {} amb {} ( {} {} {})'
        #                .format(ts,
        #                        tl.green[-1].max_ts,
        #                        tl.red[-1].max_ts,
        #                        tl.amber[-1].max_ts,
        #                        (ts - tl.green[-1].min_ts) / 1e6,
        #                        (ts - tl.red[-1].min_ts) / 1e6,
        #                        (ts - tl.amber[-1].min_ts) / 1e6))

        return tl.get_green2(ts), tl.get_red2(ts), tl.get_amber2(ts)

