from shapely.geometry import polygon, point

from edge_bbox.types import RectF, QuadI
from ...model.types import SceneArgs


class SceneMask:
    def __init__(self, scene_args):
        """
        :type scene_args: SceneArgs
        """
        self._polygon = polygon.Polygon(scene_args.mask)
        self._scene_iw, self._scene_ih = scene_args.mask_wh

    def check_rect(self, rect):
        """
        :type rect: RectF
        :rtype: bool
        """
        center_fx = (rect.fl + rect.fr) / 2.0 * self._scene_iw
        center_fy = (rect.ft + rect.fb) / 2.0 * self._scene_ih
        plate_center = point.Point(int(center_fx), int(center_fy))
        return self._polygon.contains(plate_center)

    def check_quad(self, quad):
        """
        :type quad: QuadI
        :rtype: bool
        """
        plate_center = point.Point(*quad.center)
        return self._polygon.contains(plate_center)


class RlSceneMask:
    STOPBAR_NONE = 0
    STOPBAR_PRE = 1
    STOPBAR_ON = 2
    STOPBAR_POST = 3

    _UP = 4
    _DOWN = 5
    _EQ = 6

    def __init__(self, scene_args):
        """
        :type scene_args: SceneArgs
        """
        self._scene_iw, self._scene_ih = scene_args.mask_wh

        self._stopbar_far = point.Point(*scene_args.mask[0]), point.Point(*scene_args.mask[1])
        self._stopbar_near = point.Point(*scene_args.mask[2]), point.Point(*scene_args.mask[3])
        self._stopbar_quad = polygon.Polygon(scene_args.mask)

    @staticmethod
    def _point_to_line_relation(point_to_check, line_start, line_end):
        """
        :type point_to_check: point.Point
        :type line_start: point.Point
        :type line_end: point.Point

        :rtype: int
        """
        y2 = line_end.y
        y1 = line_start.y

        x2 = line_end.x
        x1 = line_start.x

        x = point_to_check.x
        y = point_to_check.y

        # img coordinate y are reversed
        if y < (y2 - y1) / (x2 - x1) * (x - x1) + y1:
            return RlSceneMask._UP
        elif y == (y2 - y1) / (x2 - x1) * (x - x1) + y1:
            return RlSceneMask._EQ
        else:
            return RlSceneMask._DOWN

    def check_rect(self, center_lt, receding_direction=True):
        """
        :type center_lt: (float, float)
        :type receding_direction: bool
        :rtype: int
        """
        center_pt = point.Point(int(self._scene_iw * center_lt[0]), int(self._scene_ih * center_lt[1]))

        if self._stopbar_quad.contains(center_pt):
            return self.STOPBAR_ON

        elif self._point_to_line_relation(center_pt, *self._stopbar_far) == self._UP:
            return self.STOPBAR_POST if receding_direction else self.STOPBAR_PRE

        elif self._point_to_line_relation(center_pt, *self._stopbar_near) == self._DOWN:
            return self.STOPBAR_PRE if receding_direction else self.STOPBAR_POST

    def check_quad(self, quad, receding_direction=True):
        """
        :type quad: QuadI
        :type receding_direction: bool
        :rtype: int
        """
        center_pt = point.Point(*quad.center)

        if self._stopbar_quad.contains(center_pt):
            return self.STOPBAR_ON

        elif self._point_to_line_relation(center_pt, *self._stopbar_far) == self._UP:
            return self.STOPBAR_POST if receding_direction else self.STOPBAR_PRE

        elif self._point_to_line_relation(center_pt, *self._stopbar_near) == self._DOWN:
            return self.STOPBAR_PRE if receding_direction else self.STOPBAR_POST
