import math
import time
from datetime import datetime, timedelta

from edge import eventlog
from edge_bbox.types import RectF, mk_plate_bbox2, RectI, QuadI

from edge.image.image import ImEnv, ImEngine
from edge.image.types import ImFooterVars, MemBlob
from edge.service.tz import ITzService
from edge.service.about import IAboutService
from edge.model.track import LrTrackItem
from edge.model.trend import LrTrend, ImPlace
from edge.lane_detector import LaneDetector

from .types import DetectorArgs, RuleArgs, ParkingEvidence
from ..types import IDetector
from ..com.scene import SceneMask
from ..com.schedule import Schedule
from ...model.types import BaseEvent, Event, ViolationWaypoint
from ...model.models import Track
from ...services.db import IDbService
from ...services.vs import VsImageInfo, IVsService, VsImageBufferError


def create(*args, **kwargs):
    """
    :rtype: IDetector
    """
    return _Detector(*args, **kwargs)


def get_trace_duration(trace):
    """
    :type trace: list[LrTrackItem]
    :rtype: long
    """
    return trace[-1].ts - trace[0].ts


def check_rects(rect1, rect2):
    """
    :type rect1: RectF
    :type rect2: RectF
    :rtype: bool
    """
    podgonian = 2  # TODO
    rect1_c_y = (rect1.ft + rect1.fb) / 2
    rect1_c_x = (rect1.fr + rect1.fl) / 2

    rect2_c_y = (rect2.ft + rect2.fb) / 2
    rect2_c_x = (rect2.fr + rect2.fl) / 2

    dist = math.sqrt(math.pow(rect1_c_x - rect2_c_x, 2) + math.pow(rect1_c_y - rect2_c_y, 2))
    rect_1_h = rect1.fb - rect1.ft
    return dist < podgonian * rect_1_h


class _SceneMsRef(VsImageInfo):
    def __init__(self, lr_item):
        """
        :type lr_item: LrTrackItem
        """
        VsImageInfo.__init__(self, lr_item.ts)
        self.lr_item = lr_item


class _RuleServices:
    def __init__(self, tz_service, vs_service, about_service, lane_service):
        """
        :type tz_service: ITzService
        :type vs_service: IVsService
        :type about_service: IAboutService
        :type lane_service: LaneDetector
        """
        self.tz = tz_service
        self.vs = vs_service
        self.about = about_service
        self.lane = lane_service


class Rule:
    def __init__(self, detector_id, detector_name, services, args, init_im_env, fini_im_env):
        """
        :type detector_id: int
        :type detector_name: str
        :type services: _RuleServices
        :type args: RuleArgs
        :type init_im_env: ImEnv
        :type fini_im_env: ImEnv
        """
        self._detector_id = detector_id
        self._detector_name = detector_name
        self._services = services
        self._args = args

        self._init_im_env = init_im_env
        self._fini_im_env = fini_im_env

        self._schedule = Schedule(args.schedule, services.tz)

    def check(self, track, base, trend):
        """
        :type track: Track
        :type base: BaseEvent
        :type trend: LrTrend
        :rtype ParkingEvidence | None
        """
        for item in trend:
            # check constraints
            if item.event:
                continue

            elif (item.range.duration / 1e6) < self._args.limits.max_standing_time:
                continue

            elif not self._schedule.check2(item.range.min_ts):
                continue

            elif not self._schedule.check2(item.range.max_ts):
                continue

            # load init-image
            try:
                im_info = self._services.vs.export_image_rgba(
                    self._init_im_env.scene_rgba, [VsImageInfo(item.range.min_ts)])
            except VsImageBufferError as ex:
                self._init_im_env.scene_rgba.reset(14, ex.min_w, ex.min_h, self._init_im_env.footer_h)
                im_info = self._services.vs.export_image_rgba(
                    self._init_im_env.scene_rgba, [VsImageInfo(item.range.min_ts)])

            if im_info is None:
                eventlog.log_d('fail: load init-image')
                continue

            # load fini-image
            try:
                im_info = self._services.vs.export_image_rgba(
                    self._fini_im_env.scene_rgba, [VsImageInfo(item.range.max_ts)])
            except VsImageBufferError as ex:
                self._fini_im_env.scene_rgba.reset(14, ex.min_w, ex.min_h, self._fini_im_env.footer_h)
                im_info = self._services.vs.export_image_rgba(
                    self._fini_im_env.scene_rgba, [VsImageInfo(item.range.max_ts)])

            if im_info is None:
                eventlog.log_d('fail: load fini-image')
                continue

            plate_rect = self._mk_plate_rect(
                item.place,
                track.lr_track.best_image.width,
                track.lr_track.best_image.height,
                trend.plate_w, trend.plate_h)

            evidence = self._mk_evidence(base, track, plate_rect, plate_rect)
            if evidence:
                return evidence

    @staticmethod
    def _mk_plate_rect(place, scene_w, scene_h, plate_w, plate_h):
        """
        :type place: ImPlace
        :type scene_w: int
        :type scene_h: int
        :type plate_w: int
        :type plate_h: int
        :rtype: RectF
        """
        wh_ratio = float(plate_w) / plate_h
        x, y = place.xy
        r = place.r

        fl = (x - r * wh_ratio / 2.0) / float(scene_w)
        fr = (x + r * wh_ratio / 2.0) / float(scene_w)

        ft = (y - r / 2.0) / float(scene_h)
        fb = (y + r / 2.0) / float(scene_h)

        return RectF(fl, ft, fr, fb)

    def _mk_evidence(self, base_event, track, init_rect, fini_rect):
        """
        :type base_event: BaseEvent
        :type track: Track
        :type init_rect: RectF
        :type fini_rect: RectF
        :rtype ParkingEvidence | None
        """
        plate_text = base_event.get_plate_text()

        def mk_waypoint_blobs(env, plate_rect, origin):
            """
            :type env: ImEnv
            :type plate_rect: RectF
            :type origin: Origin
            :rtype: list[MemBlob]
            """
            plate_bbox = mk_plate_bbox2(plate_rect, *env.scene_rgba.wh)

            # todo: workaround
            l = plate_bbox.fl
            t = plate_bbox.ft
            r = plate_bbox.fr
            b = plate_bbox.fb
            quad = QuadI(l, t, r, t, r, b, l, b)

            lane_set = self._services.lane.get_lane3((l + r) / 2, (t + b) / 2)
            if len(lane_set) == 1:
                origin.location.lane = lane_set.pop()

            footer_vars = ImFooterVars(self._services.tz, self._services.about, env.scene_rgba.ts, env.footer_def_vars,
                                       location=origin.location, geolocation=origin.geolocation, plate_text=plate_text)

            blobs = env.engine.mk_blobs(
                env.scene_rgba, plate_bbox, footer_vars, track.lr_track, quad,
                vehicle=True, scene=True, plate=True)

            return blobs

        init_ts = self._init_im_env.scene_rgba.ts
        init_origin = self._services.about.get_origin(init_ts)
        init_blobs = mk_waypoint_blobs(self._init_im_env, init_rect, init_origin)
        if len(init_blobs) != 3:
            eventlog.log_d('fail: mk-waypoint-blobs <violation>')
            return None

        fini_ts = self._fini_im_env.scene_rgba.ts
        fini_origin = self._services.about.get_origin(fini_ts)
        fini_blobs = mk_waypoint_blobs(self._fini_im_env, fini_rect, fini_origin)
        if len(fini_blobs) != 3:
            eventlog.log_d('fail: mk-waypoint-blobs <violation>')
            return None

        standing_time = int((fini_ts - init_ts) / 1e6)
        evidence = ParkingEvidence(
            self._detector_name, self._args, fini_ts,
            self._services.about.get_origin(fini_ts),
            standing_time)

        init_rect_i = RectI.create(init_rect, track.lr_track.best_image.width, track.lr_track.best_image.height)
        fini_rect_i = RectI.create(fini_rect, track.lr_track.best_image.width, track.lr_track.best_image.height)

        evidence.add_waypoints([
            ViolationWaypoint(init_ts, init_origin.location, plate_rects=[init_rect_i]).add_images(init_blobs),
            ViolationWaypoint(fini_ts, fini_origin.location, plate_rects=[fini_rect_i]).add_images(fini_blobs)
        ])

        return evidence


class _Detector(IDetector):
    def __init__(self, tz_service, vs_service, about_service, db_service, lane_service, args):
        """
        :type tz_service: ITzService
        :type vs_service: IVsService
        :type about_service: IAboutService
        :type args: DetectorArgs
        :type db_service: IDbService
        :type lane_service: LaneDetector
        """
        # range tout
        # lost tout
        # max standing time

        IDetector.__init__(self)

        self._identity = args.identity

        self._scene_mask = SceneMask(args.scene)
        self._log_range_tout = args.range_tout
        self._log_lost_tout = args.lost_tout

        self._db_service = db_service
        self._handler = None

        rule_services = _RuleServices(tz_service, vs_service, about_service, lane_service)

        self._init_im_env = ImEnv(
            ImEngine(args.init_im_engine),
            args.init_im_engine.scene.footer.height,
            args.def_footer_mapping)

        self._fini_im_env = ImEnv(
            ImEngine(args.fini_im_engine),
            args.fini_im_engine.scene.footer.height,
            args.def_footer_mapping)

        self._rules = map(
            lambda rule_args: Rule(
                self._identity.id, args.name, rule_services, rule_args,
                self._init_im_env, self._fini_im_env),
            args.rules)

    def set_handler(self, event_handler):
        """
        :type event_handler: (Event) -> None
        """
        if self._handler:
            raise NotImplementedError

        self._handler = event_handler

    def analyze(self, track, base):
        """
        :type track: Track
        :type base: BaseEvent
        """
        if '*' in base.get_plate_text():
            eventlog.log_d('done: skip event due to asterisk in plate-text')
            return

        area_items = filter(
            lambda lr_item: self._scene_mask.check_quad(lr_item.plate_quad),
            track.lr_track)

        if not area_items:
            return

        # 1 load logs with plate-dist = 1
        # 1.1 if the is an result-entry with dist == 0, prolong it
        # 1.2 else create the new one
        # 2 try to merge result-entries

        answer = self._db_service.get_log(self._identity.id, base.get_plate_text(), plate_dist=0)
        # print 'db.get-log: %s, log-count=%s' % (base.get_plate_text(), len(log_jnodes))
        trend = LrTrend.deserialize(answer[0][2]) if answer else LrTrend()
        trend.prolong(area_items, self._log_range_tout, self._log_lost_tout)
        trend_copy = trend.retrend()

        for rule in self._rules:
            evidence = rule.check(track, base, trend_copy)
            if evidence:
                trend.set_event(evidence.ts)
                self._handler(Event(self._identity, base, evidence))
                break

        self._db_service.set_log(
            self._identity.id, base.get_plate_text(),
            trend.min_ts, trend.max_ts, trend.pack())

        # ts = (time.time() + 60) * 1e6  # (now + 1 min) in microseconds
        # max_ts = int(ts - self._log_lost_tout)
        max_datetime = datetime.now() - timedelta(days=1)
        max_ts = int(time.mktime(max_datetime.timetuple()) * 1e6)

        self._db_service.clear_log(self._identity.id, max_ts)
