import functools
import math

from edge import eventlog
from edge_bbox.types import mk_plate_bbox2, RectF
from edge.image.image import ImEnv, ImEngine, ImSizeFitter
from edge.image.types import FileBlob, ImFooterVars
from edge.service.tz import ITzService
from edge.service.about import IAboutService
from edge.model.types import TimeSpan
from edge.lane_detector import LaneDetector

from .types import DetectorArgs, RedLightEvidence, RuleArgs
from ..base.best import check_quad_dist_from_img_borders
from ..types import IDetector
from ..com.scene import RlSceneMask
from ..com.types import VideoArgs
from ...model.types import BaseEvent, Event, Waypoint, RedLightViolationWaypoint, BestWaypoint, Vehicle
from ...model.models import Track, TrTrackItem
from ...services.vs import IVsService, VsImageInfo, VsError, VsSessionAbsentError, VsImageBufferError
from ...services.tl import ITlService
from ...services.leddar import ILeddarService


def points_far_enought(rect1, rect2):
    """
    :type rect1: RectF
    :type rect2: RectF
    :rtype: bool
    """
    plate_h = rect1.fb - rect1.ft
    max_distance = plate_h * 0.7  # % of plate height

    rect1_c_x, rect1_c_y = rect1.center
    rect2_c_x, rect2_c_y = rect2.center

    if math.fabs(rect1_c_x - rect2_c_x) < max_distance and math.fabs(rect1_c_y - rect2_c_y) < max_distance:
        # points are in same rect
        return False
    else:
        # points are far enought
        return True


class _SceneImageInfo(VsImageInfo):
    def __init__(self, tr_item):
        """
        :type tr_item: TrTrackItem
        """
        VsImageInfo.__init__(self, tr_item.lr_item.ts)
        self.tr_item = tr_item


class _RuleServices:
    def __init__(self, tz_service, vs_service, about_service, tl_service, leddar_service, lane_service):
        """
        :type tz_service: ITzService
        :type vs_service: IVsService
        :type about_service: IAboutService
        :type tl_service: ITlService
        :type leddar_service: ILeddarService
        :type lane_service: LaneDetector
        """
        self.tz = tz_service
        self.vs = vs_service
        self.about = about_service
        self.tl = tl_service
        self.leddar = leddar_service
        self.lane = lane_service


class Rule:
    def __init__(self, twisted_reactor, services, rule_args, detector_name, init_im_env, postbar_init_im_env,
                 postbar_fini_im_env, video_conf):
        """
        :type twisted_reactor: twisted.internet.interfaces.IReactorFromThreads
        :type services: _RuleServices
        :type rule_args: RuleArgs
        :type detector_name: str
        :type init_im_env: ImEnv
        :type postbar_init_im_env: ImEnv
        :type postbar_fini_im_env: ImEnv
        :type video_conf: VideoArgs | None
        """
        self._reactor = twisted_reactor
        self._detector_name = detector_name

        self._services = services
        self._args = rule_args
        self._pre_im_env = init_im_env
        self._post_im_env1 = postbar_init_im_env
        self._post_im_env2 = postbar_fini_im_env

        self._video_conf = video_conf
        if video_conf:
            self._im_fitter = ImSizeFitter((video_conf.max_w, video_conf.max_h))

    def check(self, track, base, pre_stopbar, on_stopbar, post_stopbar):
        """
        :type track: Track
        :type base: BaseEvent
        :type pre_stopbar: list[TrTrackItem]
        :type on_stopbar: list[TrTrackItem]
        :type post_stopbar: list[TrTrackItem]
        :rtype: (Evidence, int) | None
        """
        tl_result = self._services.tl.check(track.lr_track.min_ts, track.lr_track.max_ts)

        leddar_result = self._services.leddar.check(track.lr_track.min_ts, track.lr_track.max_ts) # TODO
        for violation in tl_result:

            # check pre-stopbar conditions
            pre_ts_range = violation.intersect(
                TimeSpan(pre_stopbar[0].ts, pre_stopbar[-1].ts))
            if pre_ts_range is None:
                continue

            pre_tr_items = filter(
                lambda tr_item: tr_item.ts in pre_ts_range, pre_stopbar)
            if not pre_tr_items:
                continue

            # check on-stopbar conditions
            on_ts_range = violation.intersect(
                TimeSpan(on_stopbar[0].ts, on_stopbar[-1].ts))
            if on_ts_range is None:
                continue

            on_tr_items = filter(
                lambda tr_item: tr_item.ts in on_ts_range, on_stopbar)
            if not on_tr_items:
                continue

            # check post-stopbar conditions
            post_ts_range = violation.intersect(
                TimeSpan(post_stopbar[0].ts, post_stopbar[-1].ts))
            if post_ts_range is None:
                continue

            post_tr_items = filter(
                lambda tr_item: tr_item.ts in post_ts_range, post_stopbar)
            if not post_tr_items:
                continue

            # prepare im-environments
            pre_im_infos = map(_SceneImageInfo, pre_tr_items)
            post_im_infos = map(_SceneImageInfo, post_tr_items)

            try:
                pre_im_info = self._services.vs.export_image_rgba(
                    self._pre_im_env.scene_rgba, pre_im_infos[:-6:-1])  # type: _SceneImageInfo
            except VsImageBufferError as ex:
                self._pre_im_env.scene_rgba.reset(14, ex.min_w, ex.min_h, self._pre_im_env.footer_h)
                pre_im_info = self._services.vs.export_image_rgba(
                    self._pre_im_env.scene_rgba, pre_im_infos[:-6:-1])  # type: _SceneImageInfo

            if pre_im_info is None:
                eventlog.log_d('fail: load pre-stopbar image')
                continue

            try:
                post_im_info1 = self._services.vs.export_image_rgba(
                    self._post_im_env1.scene_rgba, post_im_infos[0:5])  # type: _SceneImageInfo
            except VsImageBufferError as ex:
                self._post_im_env1.scene_rgba.reset(14, ex.min_w, ex.min_h, self._post_im_env1.footer_h)
                post_im_info1 = self._services.vs.export_image_rgba(
                    self._post_im_env1.scene_rgba, post_im_infos[0:5])  # type: _SceneImageInfo

            if post_im_info1 is None:
                eventlog.log_d('fail: load post-stopbar image-1')
                continue
            try:
                post_im_info2 = self._services.vs.export_image_rgba(
                    self._post_im_env2.scene_rgba, post_im_infos[:-6:-1])  # type: _SceneImageInfo
            except VsImageBufferError as ex:
                self._post_im_env2.scene_rgba.reset(14, ex.min_w, ex.min_h, self._post_im_env2.footer_h)
                post_im_info2 = self._services.vs.export_image_rgba(
                    self._post_im_env2.scene_rgba, post_im_infos[:-6:-1])  # type: _SceneImageInfo

            if post_im_info2 is None:
                eventlog.log_d('fail: load post-stopbar image-2')
                continue

            # prepare evidence
            evidence = self._mk_evidence(
                base, track, pre_im_info.tr_item,
                post_im_info1.tr_item, post_im_info2.tr_item)

            if evidence:
                ref_ts = post_tr_items[0].lr_item.ts / 1000
                return evidence, ref_ts

    def add_video(self, ref_ts, callback):
        """
        :type ref_ts: float
        :type callback: (FileBlob | None) -> None
        :rtype: None
        """

        if self._video_conf:
            scaled_w, scaled_h = self._im_fitter.scaledown(
                self._pre_im_env.scene_rgba.wh)

            # noinspection PyBroadException
            try:
                return self._services.vs.export_video(
                    ref_ts - self._video_conf.pre_record,
                    ref_ts + self._video_conf.post_record,
                    scaled_w, scaled_h, self._video_conf.quality,
                    lambda *args, **kw: self._reactor.callFromThread(callback, *args, **kw))

            except VsSessionAbsentError:
                eventlog.log_w('fail: vs (session-absent)')

            except VsError:
                eventlog.log_e('fail: vs')

            except Exception:
                eventlog.log_x()

        callback(None)

    def _mk_evidence(self, base_event, track, pre_tr_item, post_tr_item1, post_tr_item2):
        """
        :type base_event: BaseEvent
        :type track: Track
        :type pre_tr_item: TrTrackItem
        :type post_tr_item1: TrTrackItem
        :type post_tr_item2: TrTrackItem
        :rtype: RedLightEvidence | None
        """
        plate_text = base_event.get_plate_text()

        # get best vehicle
        best_waypoint = base_event.measurements[0].waypoints[2]  # type: BestWaypoint
        best_vehicle = best_waypoint.vehicle  # type: Vehicle

        def mk_waypoint(services, im_env, tr_item, waypoint_name):
            """
            :type services: _RuleServices
            :type im_env: ImEnv
            :type tr_item: TrTrackItem
            :type waypoint_name: str
            :rtype: Waypoint
            """
            lr_scene_img = track.lr_track.best_image
            plate_rect = RectF(*tr_item.lr_item.plate_quad.get_rel_bbox(lr_scene_img.width, lr_scene_img.height))
            plate_bbox = mk_plate_bbox2(plate_rect, *im_env.scene_rgba.wh)
            origin = services.about.get_origin(im_env.scene_rgba.ts)
            wp_direction = track.tr_track.get_item_direction(tr_item)

            lane_set = self._services.lane.get_lane3(*tr_item.get_plate_quad().center)
            if len(lane_set) == 1:
                origin.location.lane = lane_set.pop()

            footer_vars = ImFooterVars(services.tz, self._services.about, im_env.scene_rgba.ts, im_env.footer_def_vars,
                                       location=origin.location, geolocation=origin.geolocation, plate_text=plate_text,
                                       spot_speed=tr_item.ra_item.speed if tr_item.ra_item else None,
                                       direction=wp_direction,
                                       radar_speed=best_vehicle.spot_speed, video_speed=best_vehicle.camera_speed,
                                       video_speed2=best_vehicle.video_speed2)

            im_blobs = im_env.engine.mk_blobs(
                im_env.scene_rgba, plate_bbox, footer_vars, track.lr_track, tr_item.lr_item.plate_quad,
                vehicle=False, scene=True, plate=True)

            if len(im_blobs) != 2:
                eventlog.log_d('fail: mk image-blobs')
                raise RuntimeError

            waypoint = RedLightViolationWaypoint(
                waypoint_name, im_env.scene_rgba.ts, origin.location, [tr_item.lr_item.plate_quad],
                *services.tl.get_times(im_env.scene_rgba.ts))

            waypoint.add_images(im_blobs)
            return waypoint

        try:
            pre_waypoint = mk_waypoint(
                self._services, self._pre_im_env, pre_tr_item,
                Waypoint.NAME_PRE_STOPBAR)

            post_waypoint1 = mk_waypoint(
                self._services, self._post_im_env1, post_tr_item1,
                Waypoint.NAME_POST_STOPBAR)

            post_waypoint2 = mk_waypoint(
                self._services, self._post_im_env2, post_tr_item2,
                Waypoint.NAME_POST_STOPBAR)

        except RuntimeError:
            return None

        event_ts = self._post_im_env1.scene_rgba.ts

        evidence = RedLightEvidence(
            self._services.about.get_origin(event_ts),
            event_ts, self._args, self._detector_name)
        evidence.add_waypoints([pre_waypoint, post_waypoint1, post_waypoint2])
        return evidence


def create(*args, **kwargs):
    """
    :rtype: IDetector
    """
    return _Detector(*args, **kwargs)


class _Detector(IDetector):
    def __init__(self, twisted_reactor, tz_service, vs_service, about_service, tl_service, leddar_service, lane_service, args):
        """
        :type twisted_reactor: twisted.internet.interfaces.IReactorFromThreads
        :type vs_service: IVsService
        :type tz_service: ITzService
        :type tl_service: ITlService
        :type leddar_service: ILeddarService
        :type lane_service: LaneDetector
        :type args: DetectorArgs
        """
        IDetector.__init__(self)

        self._sc_detector = RlSceneMask(args.scene)

        im_engine = ImEngine(args.im_engine)
        prebar_im_env = ImEnv(
            im_engine, args.im_engine.scene.footer.height, args.def_footer_mapping)

        postbar_init_im_env = ImEnv(
            im_engine, args.im_engine.scene.footer.height, args.def_footer_mapping)

        postbar_fini_im_env = ImEnv(
            im_engine, args.im_engine.scene.footer.height, args.def_footer_mapping)

        self._rules = []

        rule_services = _RuleServices(tz_service, vs_service, about_service, tl_service, leddar_service, lane_service)
        for rule in args.rules:
            rule = Rule(
                twisted_reactor,
                rule_services, rule, args.name, prebar_im_env, postbar_init_im_env,
                postbar_fini_im_env, args.video)
            self._rules.append(rule)

        self._identity = args.identity
        self._handler = None
        self._receding_direction = args.receding_direction

    def set_handler(self, event_handler):
        """
        :type event_handler: (Event) -> None
        """
        if self._handler:
            raise NotImplementedError

        self._handler = event_handler

    def handle_evidence(self, base, evidence, video_blob):
        """
        :type base: BaseEvent
        :type evidence: RedLightEvidence
        :type video_blob: FileBlob | None
        """
        if video_blob:
            evidence.add_videos([video_blob])

        self._handler(Event(self._identity, base, evidence))

    def analyze(self, track, base):
        """
        :type track: Track
        :type base: BaseEvent
        """

        def split_track(sc_detector):
            """
            :type sc_detector: RlSceneMask
            :rtype: (list[TrTrackItem], list[TrTrackItem], list[TrTrackItem]) | None
            """
            lr_scene_img = track.lr_track.best_image
            tr_items = filter(
                lambda item: item.lr_item and check_quad_dist_from_img_borders(
                    item.lr_item.plate_quad, lr_scene_img.width, lr_scene_img.height),
                track.tr_track)

            pre_stopbar, on_stopbar, post_stop_bar = [], [], []
            for tr_item in tr_items:
                check_answer = sc_detector.check_quad(tr_item.lr_item.plate_quad, self._receding_direction)

                if check_answer == sc_detector.STOPBAR_PRE:
                    pre_stopbar.append(tr_item)

                elif check_answer == sc_detector.STOPBAR_ON:
                    if len(pre_stopbar):
                        on_stopbar.append(tr_item)

                    else:
                        # we didn't have pre-stopbar items yet
                        # car direction is wrong or we din't have points before stop bar
                        # print 'err in'
                        return None

                elif check_answer == sc_detector.STOPBAR_POST:
                    if len(pre_stopbar) == 0 or len(on_stopbar) == 0:
                        # we didn't have pre-stop-bar or inside-stop-bar items yet
                        # car direction is wrong or we din't have points before stop bar
                        # print 'err post'
                        return None

                    post_stop_bar.append(tr_item)

            return pre_stopbar, on_stopbar, post_stop_bar

        if '*' in base.get_plate_text():
            eventlog.log_d('done: skip event due to asterisk in plate-text')
            return

        answer = split_track(self._sc_detector)
        if answer is None:
            # controlling direction of movement is not defined
            return

        elif 0 in map(len, answer):
            # stopbar crossing is not detected
            return

        for rule in self._rules:
            reply = rule.check(track, base, *answer)
            if reply is None:
                continue

            evidence, ref_ts = reply
            if evidence:
                handler = functools.partial(self.handle_evidence, base, evidence)
                rule.add_video(ref_ts, handler)
                break
