from uuid import UUID

from edge import eventlog
from edge.image.jnode.types import JsonImEngineArgs
from edge.model.types import Dir
from edge.types import JnodeError


from ..types import (DetectorArgs, RuleArgs, LimitsArgs,
                     SectionArgs, InitAreaArgs, FiniAreaArgs)
from ...com.jnode.types import JsonBaseRuleArgs
from ...types import IDetector
from ....model.types import Identity
from ....model.jnode.types import JsonMask


class JsonLimitsArgs(LimitsArgs):
    def __init__(self, *args, **kwargs):
        LimitsArgs.__init__(self, *args, **kwargs)

    @staticmethod
    def deserialize(limits_jnode):
        """
        :type limits_jnode: dict
        :rtype: JsonLimitsArgs
        :raise JnodeError:
        """
        # deserialize #/speed-limit
        speed_limit = limits_jnode.get('speed-limit')
        if not (int == type(speed_limit) and 0 <= speed_limit <= 250):
            eventlog.log_e_param('#/speed-limit')
            raise JnodeError

        # deserialize #/max-overspeed
        max_overspeed = limits_jnode.get('max-overspeed')
        if not (int == type(speed_limit) and 0 <= max_overspeed <= 250):
            eventlog.log_e_param('#/max-overspeed')
            raise JnodeError

        return JsonLimitsArgs(speed_limit, max_overspeed)


class JsonRuleArgs(RuleArgs):
    def __init__(self, *args, **kwargs):
        RuleArgs.__init__(self, *args, **kwargs)

    @staticmethod
    def deserialize(rule_jnode):
        """
        :type rule_jnode: dict
        :rtype: JsonRuleArgs
        :raise JnodeError:
        """
        if type(rule_jnode) != dict:
            eventlog.log_e_param('#/')
            raise JnodeError

        # deserialize #/name
        # deserialize #/enabled
        # deserialize #/schedule
        # deserialize #/ticket
        rule = JsonBaseRuleArgs.deserialize(rule_jnode)

        # deserialize #/limits
        limits_jnode = rule_jnode.get('limits')
        if type(rule_jnode) != dict:
            eventlog.log_e_param('#/limits')
            raise JnodeError

        limits = JsonLimitsArgs.deserialize(limits_jnode)

        return JsonRuleArgs(
            rule.name, rule.enabled, rule.schedule, rule.ticket,
            limits)


class JsonInitAreaArgs(InitAreaArgs):
    def __init__(self, *args, **kwargs):
        InitAreaArgs.__init__(self, *args, **kwargs)

    @staticmethod
    def deserialize(init_area_jnode):
        """
        :type init_area_jnode: dict
        :rtype: JsonInitAreaArgs
        :raise JnodeError:
        """
        if type(init_area_jnode) != dict:
            eventlog.log_e_param('#/')
            raise JnodeError

        # deserialize #/origin
        origin = init_area_jnode.get('origin')
        try:
            _ = UUID(origin)
        except ValueError:
            eventlog.log_e_param('#/origin')
            raise JnodeError

        # deserialize #/auth
        auth_jnode = init_area_jnode.get('auth')
        if not isinstance(auth_jnode, (list, tuple)) or len(auth_jnode) != 2:
            eventlog.log_e_param('#/auth')
            raise JnodeError

        user, pwd = auth_jnode
        if not (isinstance(user, (str, unicode)) and len(user) >= 1):
            eventlog.log_e_param('#/auth/0')
            raise JnodeError

        if not (isinstance(pwd, (str, unicode)) and len(pwd) >= 6):
            eventlog.log_e_param('#/auth/1')
            raise JnodeError

        # deserialize #/controller
        controller = init_area_jnode.get('controller')
        if not (controller and isinstance(controller, (str, unicode))):
            eventlog.log_e_param('#/controller')
            raise JnodeError

        return JsonInitAreaArgs(controller, (user, pwd), origin)


class JsonFiniAreaArgs(FiniAreaArgs):
    def __init__(self, *args, **kwargs):
        FiniAreaArgs.__init__(self, *args, **kwargs)

    @staticmethod
    def deserialize(fini_area_jnode):
        """
        :type fini_area_jnode: dict
        :rtype: JsonFiniAreaArgs
        :raise JnodeError:
        """
        if type(fini_area_jnode) != dict:
            eventlog.log_e_param('#/')
            raise JnodeError

        # deserialize #/mask
        mask_jnode = fini_area_jnode.get('mask')
        mask = JsonMask.deserialize_list(mask_jnode)

        # deserialize #/scene-wh
        scene_wh_jnode = fini_area_jnode.get('scene-wh')
        if not isinstance(scene_wh_jnode, (list, tuple)) or len(scene_wh_jnode) != 2:
            eventlog.log_e_param('#/scene-wh')
            raise JnodeError

        w, h = scene_wh_jnode
        if not (int == type(w) and 640 <= w <= 65535):
            eventlog.log_e_param('#/scene-wh/0')
            raise JnodeError

        if not (int == type(h) and 480 <= h <= 65535):
            eventlog.log_e_param('#/scene-wh/1')
            raise JnodeError

        return JsonFiniAreaArgs(mask, (w, h))


class JsonSectionArgs(SectionArgs):
    DIRECTIONS = ['approaching', 'receding']

    def __init__(self, *args, **kwargs):
        SectionArgs.__init__(self, *args, **kwargs)

    @staticmethod
    def deserialize(section_jnode):
        """
        :type section_jnode: dict
        :rtype: JsonSectionArgs
        :raise JnodeError:
        """
        if type(section_jnode) != dict:
            eventlog.log_e_param('#/')
            raise JnodeError

        # deserialize #/dir
        direction_jnode = section_jnode.get('dir')
        try:
            direction = Dir.deserialize(direction_jnode)
        except ValueError:
            direction = None

        if not direction or direction_jnode not in JsonSectionArgs.DIRECTIONS:
            eventlog.log_e_param('#/dir')
            raise JnodeError

        # deserialize #/dist
        dist = section_jnode.get('dist')
        if not (isinstance(dist, int) and dist >= 300):
            eventlog.log_e_param('#/dist')
            raise JnodeError

        # deserialize #/init-area
        init_area_jnode = section_jnode.get('init-area')
        if not isinstance(init_area_jnode, dict):
            eventlog.log_e_param('#/init-area')
            raise JnodeError

        init_area = JsonInitAreaArgs.deserialize(init_area_jnode)

        # deserialize #/fini-area
        fini_area_jnode = section_jnode.get('fini-area')
        if not isinstance(fini_area_jnode, dict):
            eventlog.log_e_param('#/fini-area')
            raise JnodeError

        fini_area = JsonFiniAreaArgs.deserialize(fini_area_jnode)

        return JsonSectionArgs(init_area, fini_area, direction, dist)


class JsonDetectorArgs(DetectorArgs):
    def __init__(self, *args, **kwargs):
        DetectorArgs.__init__(self, *args, **kwargs)

    @staticmethod
    def deserialize(jnode, index, def_footer_mapping):
        """
        :type jnode: dict
        :type index: int
        :type def_footer_mapping: dict
        :rtype: JsonDetectorArgs
        :raise JnodeError:
        """
        # deserialize #/name
        name = jnode.get('name')
        if not type(name) in [str, unicode] or len(name) > 32:
            eventlog.log_e_param('#/name')
            raise JnodeError

        # deserialize #/store
        store = jnode.get('store')
        if not (isinstance(store, (str, unicode)) and 1 <= len(store) <= 2):
            eventlog.log_e_param('#/store')
            raise JnodeError

        # deserialize #/rules
        rules_jnode = jnode.get('rules')
        if type(rules_jnode) != list:
            eventlog.log_e_param('#/rules')
            raise JnodeError

        rules = map(JsonRuleArgs.deserialize, rules_jnode)
        identity = Identity(IDetector.TYPE_SECTION_SPEED, index)

        # deserialize #/images
        images_jnode = jnode.get('images')
        if type(images_jnode) != dict:
            eventlog.log_e_param('#/images')
            raise JnodeError

        im_engine = JsonImEngineArgs.deserialize(images_jnode['fini'])

        # deserialize #/section
        section_jnode = jnode.get('section')
        if type(section_jnode) != dict:
            eventlog.log_e_param('#/section')
            raise JnodeError

        section = JsonSectionArgs.deserialize(section_jnode)

        return JsonDetectorArgs(identity, name, store, rules, im_engine,
                                def_footer_mapping, section)
