from edge.image.types import ImEngineArgs
from edge.service.tz import ITzService
from edge.model.types import Dir, Origin

from ...detectors.com.types import ScheduleArgs, VideoArgs
from ...model.types import Identity, Evidence, BestViolationWaypoint


class LimitsArgs:
    def __init__(self, lanes, speed_limit,  over_speed):
        """
        :type lanes: list of str
        :type speed_limit: int
        :type over_speed: int
        """
        self.lanes = lanes
        self.speed_limit = speed_limit
        self.over_speed = over_speed


class RuleArgs:
    def __init__(self, name, enabled, schedule, ticket, limits):
        """
        :type name: str
        :type enabled: bool
        :type schedule: ScheduleArgs | None
        :type ticket: Ticket
        :type limits: LimitsArgs
        """
        self.name = name
        self.enabled = enabled
        self.schedule = schedule
        self.ticket = ticket
        self.limits = limits


class DetectorArgs:
    def __init__(self, identity, name, store, rules, im_engine, def_footer_mapping, video_args,
                 h_over_ground=None, log=False):
        """
        :type identity: Identity
        :type name: str
        :type store: str
        :type rules: list of RuleArgs
        :type im_engine: ImEngineArgs
        :type def_footer_mapping: dict
        :type video_args: VideoArgs | None
        :type h_over_ground: float | None
        :type log: bool
        """
        self.identity = identity
        self.name = name
        self.store = store
        self.rules = rules
        self.im_engine = im_engine
        self.def_footer_mapping = def_footer_mapping
        self.video = video_args
        self.h_over_ground = h_over_ground
        self.log = log


class SpotSpeedEvidence(Evidence):
    def __init__(self, origin, ts, speed, cam_speed,  rule_args, detector_name, direction):
        """
        :type origin: Origin
        :type ts: long
        :type speed: int
        :type cam_speed: int
        :type rule_args: RuleArgs
        :type detector_name: str
        :type direction: Dir
        """
        Evidence.__init__(self, Evidence.NAME_SPOT_SPEED, origin, ts)
        self.detector_name = detector_name
        self.speed = speed
        self.cam_speed = cam_speed
        self.rule_args = rule_args
        self.dir = direction

    def serialize(self, tz_service):
        """
        :type tz_service: ITzService
        :rtype: dict
        """
        jnode = Evidence.serialize(self, tz_service)
        violation_waypoints_jnode = list()
        best_waypoints_jnode = None
        for waypoint in self._waypoints:
            if waypoint.name == waypoint.NAME_VIOLATION:
                violation_waypoints_jnode.append(waypoint.serialize(tz_service))
            elif waypoint.name == waypoint.NAME_VIOLATION_BEST:
                best_waypoints_jnode = waypoint.serialize(tz_service)

        jnode['evidence'] = {
            'detector': {
                'name': self.detector_name,
                'rule': {
                    'name': self.rule_args.name,
                    'schedule': self.rule_args.schedule.serialize(),
                    'ticket': {
                        'name': self.rule_args.ticket.name
                    },
                    'limits': {
                        'speed-limit': self.rule_args.limits.speed_limit,
                        'max-overspeed': self.rule_args.limits.over_speed
                    }
                }
            },
            'measurements': {
                'speed': {
                  'radar': self.speed,
                  'camera': self.cam_speed
                },
                'waypoints': {
                    'violation': violation_waypoints_jnode
                }
            }
        }

        if best_waypoints_jnode is not None:
            jnode['evidence']['measurements']['waypoints']['best'] = best_waypoints_jnode

        if self.dir.dir != Dir.UNKNOWN:
            jnode['evidence']['measurements']['dir'] = str(self.dir)

        if len(self._videos):
            videos_jnode = map(lambda x: x.serialize(tz_service), self._videos)
            jnode['evidence']['measurements']['videos'] = videos_jnode

        return jnode

    @property
    def best_waypoint(self):
        """
        :rtype: BestViolationWaypoint | None
        """
        bw = filter(lambda x: x.name == BestViolationWaypoint.NAME_VIOLATION_BEST, self.waypoints)
        return bw[0] if len(bw) else None
