import ctypes

from edge import eventlog
from edge.image.types import Blob, FileBlob, MemBlob, StructBlobMeta
from edge.types import BufferSizeError

from edge.com import bs

from .sign import (init_service as init_sign_service,
                   fini_service as fini_sign_service,
                   SignError)
from .types import BsServiceArgs
from ..model.types import Identity


class IBsService:
    def __init__(self):
        pass

    def save(self, sender, blob):
        """
        :type sender: Identity
        :type blob: Blob
        :raise bs.BsError:
        """
        raise NotImplementedError

    def get_size(self, sender, blob_id, profile):
        """
        :type sender: Identity
        :type profile: str
        :type blob_id: str
        :rtype: int
        """
        raise NotImplementedError

    def sign(self, blob, attach_crt=True):
        """
        :type blob: MemBlob
        :type attach_crt: bool
        """
        raise NotImplementedError


def init_service(args, sign_args, counters):
    """
    :type args: BsServiceArgs
    :type sign_args: SignServiceArgs | None
    :type counters: bs.IBsCounters
    :rtype: IBsService
    """
    return _BsService(args, sign_args, counters)


def fini_service(service):
    """
    :type service: IBsService | _BsService
    """
    service.destroy()


class _BsService(IBsService):
    def __init__(self, bs_conf, sign_conf, counters):
        """
        :type bs_conf: BsServiceArgs
        :type sign_conf: SignServiceArgs | None
        :type counters: bs.IBsCounters
        """
        IBsService.__init__(self)
        self._session_args = counters, 65535, bs.LOG_LEVEL_NAME_TO_CODE_MAP.get(bs_conf.log_level, 'info')
        self._session = bs.Session(counters=counters, closed=True)

        self._counters = counters
        self._sender_to_chan_map = bs_conf.sender_to_chan_map

        self._sign_buffer_size = 1
        self._hash_buffer_size = 100

        self._sign_service = init_sign_service(sign_conf)

    def destroy(self):
        if self._sign_service:
            fini_sign_service(self._sign_service)

        self._session.close()

    def save(self, sender, blob):
        """
        :type sender: Identity
        :type blob: MemBlob | FileBlob
        """
        chan_or_profiles = self._sender_to_chan_map[sender]
        if dict == type(chan_or_profiles):
            chan = chan_or_profiles[blob.profile]
        else:
            chan = chan_or_profiles

        self._counters.inc_blob(chan)
        self.save2(chan, blob)

    def get_size(self, sender, blob_id, profile):
        """
        :type sender: Identity
        :type profile: str
        :type blob_id: str
        :rtype: int
        """
        chan_or_profiles = self._sender_to_chan_map[sender]
        if dict == type(chan_or_profiles):
            chan = chan_or_profiles[profile]
        else:
            chan = chan_or_profiles

        reply = self._session.get_blob_size(chan, blob_id)
        return reply

    @bs.reopen_session
    def save2(self, chan, blob):
        """
        :type chan: str
        :type blob: MemBlob | FileBlob

        :raise bs.BsError:
        :raise bs.BsSessionError:
        """
        try:
            blob_ts = blob.min_ts / 1000
            if blob.file_path is None:
                if isinstance(blob.meta, StructBlobMeta):
                    ident = self._session.save_blob(chan, blob_ts, blob.file_data, blob.meta.struct)
                else:
                    ident = self._session.save_blob(chan, blob_ts, blob.file_data, blob.file_meta)
            else:
                ident = self._session.save_blob2(chan, blob_ts, blob.file_path, blob.file_meta)

        except bs.BsError:
            eventlog.log_e('fail: save <chan={}, profile={}>'.format(str(chan), blob.profile))
            raise

        blob.name = '%s_%s.%s' % (chan, ident, blob.file_ext)

        hash_buffer = ctypes.create_string_buffer('', size=self._hash_buffer_size)
        hash_size = 0
        try:
            hash_size = self._sign_service.hash_blob(blob.file_data, hash_buffer)
        except BufferSizeError as x:
            self._hash_buffer_size = x.min_size
            hash_buffer = ctypes.create_string_buffer('', size=self._hash_buffer_size)
            hash_size = self._sign_service.hash_blob(blob.file_data, hash_buffer)

        except SignError:
            eventlog.log_e('fail: hash <chan={}, profile={}>'.format(str(chan), blob.profile))

        if hash_size > 0:
            blob.hash = self._sign_service.hash_type, hash_buffer.raw[:hash_size]

    def sign(self, blob, attach_crt=True):
        """
        :type blob: MemBlob
        :type attach_crt: bool
        :rtype: bool
        :returns: True if signed
        """
        sign_buffer = ctypes.create_string_buffer('', size=self._sign_buffer_size)
        try:
            sign_size = self._sign_service.sign_blob(blob._data, sign_buffer)
        except BufferSizeError as x:
            self._sign_buffer_size = x.min_size
            sign_buffer = ctypes.create_string_buffer('', size=self._sign_buffer_size)
            sign_size = self._sign_service.sign_blob(blob._data, sign_buffer)
        except SignError:
            return False

        if attach_crt:
            crt = self._sign_service.cert
            if crt:
                blob.meta = StructBlobMeta(sign_buffer.raw, sign_size, crt)
            else:
                blob.meta = StructBlobMeta(sign_buffer.raw, sign_size)
        else:
            blob.meta = StructBlobMeta(sign_buffer.raw, sign_size)

        return bool(sign_size)
