import ctypes
from glob import glob
import ujson

from .types import LandmarkServiceArgs
from .jnode.types import JsonLandmarkServiceArgs
from edge_bbox.types import QuadI

# struct camargs {
#     double    pos_l;            /* aPosL */
#     double    pos_u;            /* aPosU */
#     double    pos_f;            /* aPosF */
#     double    rot_e;            /* aRotUf */
#     double    rot_a;            /* aRotFl */
#     double    rot_r;            /* aRotLu */
#     double    near_clip_f;      /* aNearClipF */
#     double    far_clip_f;       /* aFarClipF */
#     double    sensor_w_mm;      /* aSensorWmm */
#     double    sensor_h_mm;      /* aSensorHmm */
#     int       sensor_w_px;      /* aSensorWpx */
#     int       sensor_h_px;      /* aSensorHpx */
#     int       focal_dist;       /* aFocalDist */
# };


class CamArgs(ctypes.Structure):
    _fields_ = [("pos_l", ctypes.c_double), ("pos_u", ctypes.c_double), ("pos_f", ctypes.c_double),
               ("rot_e", ctypes.c_double), ("rot_a", ctypes.c_double), ("rot_r", ctypes.c_double),
               ("near_clip_f", ctypes.c_double), ("far_clip_f", ctypes.c_double),
               ("sensor_w_mm", ctypes.c_double), ("sensor_h_mm", ctypes.c_double),
               ("sensor_w_px", ctypes.c_int), ("sensor_h_px", ctypes.c_int), ("focal_dist", ctypes.c_int)]


class QuadArgs(ctypes.Structure):
    _fields_ = [("x0", ctypes.c_int), ("y0", ctypes.c_int),
               ("x1", ctypes.c_int), ("y1", ctypes.c_int),
               ("x2", ctypes.c_int), ("y2", ctypes.c_int),
               ("x3", ctypes.c_int), ("y3", ctypes.c_int)]


class _Api:
    _lib = ctypes.CDLL(glob('/opt/edge/lib/liblandmark*.so')[0])

    # bool
    # LlMeasureSpeed(struct camargs *aCamArgs,
    #     struct quad *aQuad1, struct quad *aQuad2,
    #     double aForward1, double aForward2,
    #     uint64_t aTs1, uint64_t aTs2,
    #     double *aSpeed);

    measureSpeed = _lib.LlMeasureSpeed2
    measureSpeed.argtypes = (
        ctypes.POINTER(CamArgs), ctypes.POINTER(QuadArgs), ctypes.POINTER(QuadArgs),
        ctypes.c_double, ctypes.c_double, ctypes.c_uint64, ctypes.c_uint64,
        ctypes.POINTER(ctypes.c_double))

    measureSpeed.restype = ctypes.c_bool

    # bool
    # LlMkWorldPoint(
    #     struct camargs *aCamArgs,
    #     int anImageL,
    #     int anImageT,
    #     double anUp,
    #     double *aForward,
    #     double *aLeft
    #     )

    mkWorldPoint = _lib.LlMkWorldPoint
    mkWorldPoint.argtypes = (
        ctypes.POINTER(CamArgs), ctypes.c_int, ctypes.c_int,
        ctypes.c_double,
        ctypes.POINTER(ctypes.c_double),
        ctypes.POINTER(ctypes.c_double))

    mkWorldPoint.restype = ctypes.c_bool

    def __init__(self):
        pass


def parse_conf(conf_filepath):
    """
    :type conf_filepath: str
    :rtype: LandmarkServiceArgs
    """
    with open(conf_filepath) as json_file:
        conf_jnode = ujson.load(json_file)

    args = JsonLandmarkServiceArgs.deserialize(conf_jnode)
    return args


def get_speed2(filepath, quad1, quad2, forward1, forward2, ts1, ts2):
    """
    :type filepath: str
    :type quad1: QuadI
    :type quad2: QuadI
    :type forward1: float
    :type forward2: float
    :type ts1: int
    :type ts2: int
    :rtype: (bool, float)
    """
    args = parse_conf(filepath)

    return get_speed(args, quad1, quad2, forward1, forward2, ts1, ts2)


def get_speed(args, quad1, quad2, forward1, forward2, ts1, ts2):
    """
    :type args: LandmarkServiceArgs
    :type quad1: QuadI
    :type quad2: QuadI
    :type forward1: float
    :type forward2: float
    :type ts1: int
    :type ts2: int
    :rtype: (bool, float)
    """

    camargs = CamArgs(args.left, args.up, args.forward, args.elevation, args.azimuth, args.pitch,
                      args.clip_near, args.clip_far, args.sensor_w_mm, args.sensor_h_mm,
                      args.sensor_w_px, args.sensor_h_px, args.focal_dist)

    q1 = QuadArgs(quad1.x0, quad1.y0, quad1.x1, quad1.y1, quad1.x2, quad1.y2, quad1.x3, quad1.y3)

    q2 = QuadArgs(quad2.x0, quad2.y0, quad2.x1, quad2.y1, quad2.x2, quad2.y2, quad2.x3, quad2.y3)

    speed = ctypes.c_double()
    rc = _Api.measureSpeed(ctypes.byref(camargs), ctypes.byref(q1), ctypes.byref(q2),
         forward1, forward2, ts1, ts2, ctypes.byref(speed))

    return rc, speed.value


def get_world_coord(filepath, imLeft, imTop):
    args = parse_conf(filepath)
    camargs = CamArgs(args.left, args.up, args.forward, args.elevation, args.azimuth, args.pitch,
                      args.clip_near, args.clip_far, args.sensor_w_mm, args.sensor_h_mm,
                      args.sensor_w_px, args.sensor_h_px, args.focal_dist)

    forward = ctypes.c_double()
    left = ctypes.c_double()
    rc = _Api.mkWorldPoint(ctypes.byref(camargs), imLeft, imTop, 0.0, ctypes.byref(forward), ctypes.byref(left))
    return forward.value, left.value



def test(forward1, forward2, ts1, ts2):

    pos_l = 0
    pos_u = 5.85
    pos_f = 0
    rot_e = 9.5  #/* aRotUf */
    rot_a = 14.25  #/* aRotFl */
    rot_r = 0.0  #/* aRotLu */
    near_clip_f = 5
    far_clip_f = 80
    sensor_w_mm = 12.8
    sensor_h_mm = 9.6
    sensor_w_px = 2304
    sensor_h_px = 1728
    focal_dist = 50

    camargs = CamArgs(pos_l, pos_u, pos_f, rot_e, rot_a, rot_r, near_clip_f, far_clip_f, sensor_w_mm, sensor_h_mm,
                      sensor_w_px, sensor_h_px, focal_dist)

    q1_x0, q1_y0, q1_x1, q1_y1, q1_x2, q1_y2, q1_x3, q1_y3 = (1356, 1220, 1514, 1230, 1512, 1260, 1354, 1250)
    quad1 = QuadArgs(q1_x0, q1_y0, q1_x1, q1_y1, q1_x2, q1_y2, q1_x3, q1_y3)

    q2_x0, q2_y0, q2_x1, q2_y1, q2_x2, q2_y2, q2_x3, q2_y3 = (1510, 1147, 1641, 1155, 1639, 1186, 1508, 1178)
    quad2 = QuadArgs(q2_x0, q2_y0, q2_x1, q2_y1, q2_x2, q2_y2, q2_x3, q2_y3)

    speed = ctypes.c_double()
    rc = _Api.measureSpeed(ctypes.byref(camargs), ctypes.byref(quad1), ctypes.byref(quad2),
                           forward1, forward2, ts1, ts2, ctypes.byref(speed))

if __name__ == '__main__':
    test(22.65, 24.12, 0, 84000)



