import ctypes

from glob import glob
from edge import eventlog
from edge.types import BufferSizeError
from .types import SignServiceArgs


class SignError(RuntimeError):
    pass


class _SignSessionError(SignError):
    pass


class ISignService:
    def __init__(self):
        pass

    @property
    def cert(self):
        """
        :rtype: str
        """
        raise NotImplementedError

    @property
    def hash_type(self):
        """
        :rtype: str
        """
        raise NotImplementedError

    def get_crt(self, crt_buffer):
        """
        :type crt_buffer: str

        :rtype: int
        :raise SignError:
        """
        raise NotImplementedError

    def sign_blob(self, blob_buffer, sign_buffer):
        """
        :type blob_buffer: str
        :type sign_buffer: str

        :rtype: int
        :raise SignError:
        """
        raise NotImplementedError

    def hash_blob(self, blob_buffer, hash_buffer):
        """
        :type blob_buffer: str
        :type hash_buffer: str

        :rtype: int
        :raise SignError:
        """
        raise NotImplementedError


class _StubService(ISignService):
    def __init__(self):
        ISignService.__init__(self)

    @property
    def cert(self):
        """
        :rtype: str
        """
        return ''

    @property
    def hash_type(self):
        """
        :rtype: str
        """
        return ''

    def get_crt(self, crt_buffer):
        return 0

    def sign_blob(self, blob_buffer, sign_buffer):
        return 0

    def hash_blob(self, blob_buffer, hash_buffer):
        return 0

    def destroy(self):
        pass


def init_service(args):
    """
    :type args: SignServiceArgs | None
    :rtype: ISignService
    """
    if args is None:
        eventlog.log_i('using sign stub')
        return _StubService()

    provider = None
    provider_conf = None

    if _IitApi._lib is not None:
        provider = _IitApi
        provider_conf = args.iit_conf
        eventlog.log_i('using libsign-iit')
    elif _OpensslApi._lib is not None:
        provider = _OpensslApi
        provider_conf = args.openssl_conf
        eventlog.log_i('using libsign-openssl')

    if provider:
        log_level = provider.log_level_name_to_code(args.log_level)
        provider.init(log_level)

    return _SignService(provider_conf, provider, args) if provider else _StubService()


def fini_service(service):
    """
    :type service: ISignService | _SignService
    """
    service.destroy()


class _OpensslApi:
    LOG_N = 0
    LOG_E = 1
    LOG_W = 2
    LOG_I = 3
    LOG_D = 4
    LOG_T = 5

    RC_DONE = 0
    RC_FAIL = -1
    RC_DENY = -2

    _glob_result = glob('/opt/edge/lib/libsign.so')
    _libpath = _glob_result[0] if len(_glob_result) else None

    _lib = None
    if _libpath:
        _lib = ctypes.CDLL(_libpath)

        # int
        # EsInit(uint8_t aLogLevel)
        init = _lib.EsInit
        init.restype = ctypes.c_int
        init.argtypes = (ctypes.c_uint8,)

        # int
        # EsFini()
        fini = _lib.EsFini
        fini.argtypes = ()

        # int
        # EsOpen(const char *aCfgFile, void **aSession);
        open = _lib.EsOpen
        open.restype = ctypes.c_int
        open.argtypes = (ctypes.c_char_p, ctypes.POINTER(ctypes.c_voidp))

        # int
        # EsSign2(void *aSession, const uint8_t *aBlobData, uint32_t aBlobSize,
        #     uint8_t *aSignData, uint32_t *aSignSize);
        sign2 = _lib.EsSign2
        sign2.restype = ctypes.c_int
        sign2.argtypes = (ctypes.c_voidp, ctypes.POINTER(ctypes.c_uint8), ctypes.c_uint32,
                          ctypes.POINTER(ctypes.c_uint8), ctypes.POINTER(ctypes.c_uint32))

        # int
        # EsHash2 (void *aSession, const uint8_t *aBlobData, uint32_t aBlobSize,
        #     uint8_t *aHashData, uint32_t *aHashSize);
        hash2 = _lib.EsHash2
        hash2.restype = ctypes.c_int
        hash2.argtypes = (ctypes.c_voidp, ctypes.POINTER(ctypes.c_uint8), ctypes.c_uint32,
                          ctypes.POINTER(ctypes.c_uint8), ctypes.POINTER(ctypes.c_uint32))

        # void
        # EsClose (void *aSession);
        close = _lib.EsClose
        close.argtype = ctypes.c_voidp

    hash_type = 'md5'

    def __init__(self):
        pass

    @classmethod
    def log_level_name_to_code(cls, name):
        name_to_code = {
            'none': cls.LOG_N,
            'error': cls.LOG_E,
            'warn': cls.LOG_W,
            'info': cls.LOG_I,
            'debug': cls.LOG_D,
            'trace': cls.LOG_T
        }
        return name_to_code.get(name, cls.LOG_I)


class _IitApi:
    LOG_N = 0
    LOG_E = 1
    LOG_W = 2
    LOG_I = 0x33  # 0x13 stdout, 0x24 syslog
    LOG_D = 0x34  # 0x14 stdout, 0x23 syslog
    LOG_T = 5

    RC_DONE = 0
    RC_FAIL = -1
    RC_DENY = -2

    _glob_result = glob('/opt/edge/lib/libsigniit.so')
    _libpath = _glob_result[0] if len(_glob_result) else None

    _lib = None
    if _libpath:
        _lib = ctypes.CDLL(_libpath)

        # int
        # SiitInit(uint8_t aLogLevel)
        init = _lib.SiitInit
        init.restype = ctypes.c_int
        init.argtypes = [ctypes.c_uint8]

        # int
        # SiitFini()
        fini = _lib.SiitFini
        fini.argtypes = []

        # int
        # SiitOpen (const char *aCfgFile, void **aSession);
        open = _lib.SiitOpen
        open.restype = ctypes.c_int
        open.argtypes = (ctypes.c_char_p, ctypes.POINTER(ctypes.c_voidp))

        # int
        # SiitGetCrt (void *aSession, uint8_t *aCrtData, uint32_t *aCrtSize);
        getcrt = _lib.SiitGetCrt
        getcrt.restype = ctypes.c_int
        getcrt.argtypes = (ctypes.c_voidp, ctypes.POINTER(ctypes.c_uint8), ctypes.POINTER(ctypes.c_uint32))

        # int
        # SiitSign2 (void *aSession, const uint8_t *aBlobData, uint32_t aBlobDataSize,
        #     uint8_t *anOutSignBuff, uint32_t *anOutSignBuffSize);
        sign2 = _lib.SiitSign2
        sign2.restype = ctypes.c_int
        sign2.argtypes = (ctypes.c_voidp, ctypes.POINTER(ctypes.c_uint8), ctypes.c_uint32,
                          ctypes.POINTER(ctypes.c_uint8), ctypes.POINTER(ctypes.c_uint32))

        # int
        # SiitHash2 (void *aSession, const uint8_t *aBlobData, uint32_t aBlobDataSize,
        #     uint8_t *anOutHashBuff, uint32_t *anOutHashBuffSize);
        hash2 = _lib.SiitHash2
        hash2.restype = ctypes.c_int
        hash2.argtypes = (ctypes.c_voidp, ctypes.POINTER(ctypes.c_uint8), ctypes.c_uint32,
                          ctypes.POINTER(ctypes.c_uint8), ctypes.POINTER(ctypes.c_uint32))

        # void
        # SiitClose (void *aSession);
        close = _lib.SiitClose
        close.argtype = ctypes.c_voidp

    hash_type = 'gost34311'

    def __init__(self):
        pass

    @classmethod
    def log_level_name_to_code(cls, name):
        name_to_code = {
            'none': cls.LOG_N,
            'error': cls.LOG_E,
            'warn': cls.LOG_W,
            'info': cls.LOG_I,
            'debug': cls.LOG_D,
            'trace': cls.LOG_T
        }
        return name_to_code.get(name, cls.LOG_I)


def _check_session(fn):
    def proxy_fn(self, *args, **kwargs):
        """
        :type self: _Session
        :raise _SignSessionError:
        """
        if self.closed:
            eventlog.log_e('fail: invalid session')
            raise _SignSessionError

        return fn(self, *args, **kwargs)

    return proxy_fn


class _Session:
    def __init__(self, handle, api=None):
        """
        :type handle: ctypes.c_void_p | None
        """
        self._handle = handle
        self.closed = handle is None
        self.api = api

    @staticmethod
    def open(conf_filepath, api):
        """
        :type conf_filepath: str
        :type api: _OpensslApi | _IitApi
        :rtype: _Session

        :raise VsSessionError:
        """
        handle = ctypes.c_void_p(0)
        rc = api.open(conf_filepath, handle)
        if rc == api.RC_DENY:
            eventlog.log_e('fail: Open <RC_DENY>')
            raise SignError  # do raise _SignSessionError to retry open session
        elif rc != api.RC_DONE:
            eventlog.log_e('fail: Open <{}>'.format(rc))
            raise _SignSessionError

        return _Session(handle=handle, api=api)

    @_check_session
    def get_crt(self, crt_buffer):
        """
        :param crt_buffer: ctypes.create_string_buffer
        :type crt_buffer: str

        :rtype: int
        :raise SignError:
        """
        crt_data = ctypes.cast(crt_buffer, ctypes.POINTER(ctypes.c_uint8))

        init_crt_size = len(crt_buffer)
        crt_size = ctypes.c_uint32(init_crt_size)
        p_crt_size = ctypes.pointer(crt_size)

        rc = self.api.getcrt(self._handle, crt_data, p_crt_size)
        if rc == self.api.RC_DENY:
            eventlog.log_e('fail: GetCrt <RC_DENY>')
            raise SignError
        elif rc != self.api.RC_DONE:
            eventlog.log_e('fail: GetCrt <{}>'.format(rc))
            raise SignError

        fini_crt_size = crt_size.value
        if fini_crt_size > init_crt_size:
            eventlog.log_e('fail: crt-buffer too small (min-size: {})'.format(fini_crt_size))
            raise BufferSizeError(fini_crt_size)

        return fini_crt_size

    @_check_session
    def sign_buffer(self, blob_buffer, sign_buffer):
        """
        :type blob_buffer: str
        :param sign_buffer: ctypes.create_string_buffer
        :type sign_buffer: str

        :rtype: int
        :raise SignError:
        """
        blob_data = ctypes.cast(blob_buffer, ctypes.POINTER(ctypes.c_uint8))
        blob_size = ctypes.c_uint32(len(blob_buffer))

        sign_data = ctypes.cast(sign_buffer, ctypes.POINTER(ctypes.c_uint8))

        init_sign_size = len(sign_buffer)
        sign_size = ctypes.c_uint32(init_sign_size)
        p_sign_size = ctypes.pointer(sign_size)

        rc = self.api.sign2(self._handle, blob_data, blob_size, sign_data, p_sign_size)
        if rc == self.api.RC_DENY:
            eventlog.log_e('fail: Sign <RC_DENY>')
            raise SignError
        elif rc != self.api.RC_DONE:
            eventlog.log_e('fail: Sign <answer:{}>'.format(rc))
            raise SignError

        fini_sign_size = sign_size.value
        if fini_sign_size > init_sign_size:
            eventlog.log_e('fail: sign-buffer too small (min-size: {})'.format(fini_sign_size))
            raise BufferSizeError(fini_sign_size)

        return fini_sign_size

    @_check_session
    def hash_buffer(self, blob_buffer, hash_buffer):
        """
        :type blob_buffer: str
        :param hash_buffer: ctypes.create_string_buffer
        :type hash_buffer: str

        :rtype: int
        :raise SignError:
        """
        blob_data = ctypes.cast(blob_buffer, ctypes.POINTER(ctypes.c_uint8))
        blob_size = ctypes.c_uint32(len(blob_buffer))

        hash_data = ctypes.cast(hash_buffer, ctypes.POINTER(ctypes.c_uint8))

        init_hash_size = len(hash_buffer)
        hash_size = ctypes.c_uint32(init_hash_size)
        p_hash_size = ctypes.pointer(hash_size)

        rc = self.api.hash2(self._handle, blob_data, blob_size, hash_data, p_hash_size)
        if rc == self.api.RC_DENY:
            eventlog.log_e('fail: Hash <RC_DENY>')
            raise SignError
        elif rc != self.api.RC_DONE:
            eventlog.log_e('fail: Hash (%s)' % rc)
            raise SignError

        fini_hash_size = hash_size.value
        if fini_hash_size > init_hash_size:
            eventlog.log_e('fail: hash-buffer too small (min-size: {})'.format(fini_hash_size))
            raise BufferSizeError(fini_hash_size)

        return fini_hash_size

    def close(self):
        self.api.close(self._handle)
        self.closed = True


def _reopen_session(fn):
    def proxy_fn(self, *args, **kwargs):
        """
        :type self: _SignService
        """
        try_count = 0
        while try_count <= 3:
            try_count += 1
            try:
                if self._session.closed:
                    self._session = _Session.open(*self._session_conf)

                return fn(self, *args, **kwargs)

            except _SignSessionError:
                self._session.close()

        raise SignError

    return proxy_fn


class _SignService(ISignService):
    def __init__(self, provider_conf, provider, conf):
        """
        :type provider_conf: str
        :type provider: _OpensslApi | _IitApi
        :type conf: SignServiceArgs
        """
        ISignService.__init__(self)

        self._provider = provider
        self._hash_type = provider.hash_type
        self._session_conf = provider_conf, provider
        self._session = _Session(handle=None)
        self._conf = conf
        self._cert = None

    @property
    def cert(self):
        """
        :rtype: str
        """
        if self._cert is None:
            if self._provider == _IitApi:
                crt_buffer = ctypes.create_string_buffer('', size=1)
                try:
                    self.get_crt(crt_buffer)
                except BufferSizeError as x:
                    crt_buffer = ctypes.create_string_buffer('', size=x.min_size)
                    self.get_crt(crt_buffer)
                    self._cert = crt_buffer.raw
            else:
                self._cert = self._conf.cert

        return self._cert

    @property
    def hash_type(self):
        return self._hash_type

    @_reopen_session
    def get_crt(self, crt_buffer):
        crt_size = self._session.get_crt(crt_buffer)
        return crt_size

    def destroy(self):
        self._session.close()
        self._provider.fini()

    @_reopen_session
    def sign_blob(self, blob_buffer, sign_buffer):
        sign_size = self._session.sign_buffer(blob_buffer, sign_buffer)
        return sign_size

    @_reopen_session
    def hash_blob(self, blob_buffer, hash_buffer):
        hash_size = self._session.hash_buffer(blob_buffer, hash_buffer)
        return hash_size
