from twisted.internet import inotify
from twisted.python.filepath import FilePath

from edge import eventlog
from edge.com import fs
from edge.types import JnodeError

from .jnode.types import JsonTlServiceArgs
from edge.model.types import TimeSpan


class ITlService:
    def __init__(self):
        pass

    def check(self, min_ts, max_ts):
        """
        :type min_ts: long
        :type max_ts: long
        :rtype: list[TimeSpan]
        """
        raise NotImplementedError

    def get_times(self, ts):
        """
        :param ts: value in usec
        :type ts: long

        :return: (green-amber-time, red-time, red-amber-time). Values in usec
        :rtype: (long, long, long)
        """
        raise NotImplementedError

    def destroy(self):
        pass


def init_service(twisted_reactor, args):
    """
    :type twisted_reactor: twisted.internet.interfaces.IReactorUNIX
    :type args: JsonTlServiceArgs
    :rtype: ITlService
    """
    if args is None:
        rl = _StubService()
    else:
        rl = _TlService(twisted_reactor, args)
    return rl


def fini_service(service):
    """
    :type service: ITlService | _TlService
    """
    service.destroy()


class _StubService(ITlService):
    def __init__(self):
        ITlService.__init__(self)

    def check(self, min_ts, max_ts):
        return []

    def get_times(self, ts):
        return 0, 0, 0


def _print_ti(string, ti_l):
    """
    :type ti_l:list[TimeSpan]
    """
    print string
    for ti in ti_l:  # type: TimeSpan
        print ti.min_ts, '-', ti.max_ts


class _TrafficLight:
    def __init__(self, ch_id, red, amber, green, red_only=False):
        """
        :type ch_id: int
        :type red: list[TimeSpan]
        :type amber: list[TimeSpan]
        :type green: list[TimeSpan]
        """
        self.ch_id = ch_id
        self.red = red
        self.amber = amber
        self.green = green

        if red_only:
            self.prohibiting = red
        else:
            self.prohibiting = self._get_prohibiting()

    def get_red_time(self, ts):
        """
        :type ts: long
        :rtype: long
        """
        pr_list = filter(lambda time_span: ts in time_span, self.prohibiting)

        if not len(pr_list):
            # current collor is green
            return 0

        pr_item = pr_list[0]  # type: TimeSpan

        # find red intervals in prohibited item
        red_spans = filter(lambda span: span.check_overlap(pr_item), self.red)  # type: list[TimeSpan]

        # red started before ts
        red_spans = filter(lambda span: span.min_ts < ts, red_spans)  # type: list[TimeSpan]
        if not len(red_spans):
            return 0

        # check "just red" instead of "any red"
        # just_red = any_red - any_amber
        any_red_time = ts - red_spans[0].min_ts

        post_amber = self.get_red_amber_time(ts) or 0
        return any_red_time - post_amber

    def get_green_amber_time(self, ts):
        """
        :type ts: long
        :rtype: long
        """
        pr_list = filter(lambda time_span: ts in time_span, self.prohibiting)

        if not len(pr_list):
            # current collor is green
            return 0

        pr_item = pr_list[0]  # type: TimeSpan

        # find amber intervals in prohibited item
        amber_spans = filter(lambda span: span.check_overlap(pr_item), self.amber)  # type: list[TimeSpan]

        # amber started before ts
        amber_spans = filter(lambda span: span.min_ts < ts, amber_spans)  # type: list[TimeSpan]

        if not len(amber_spans):
            return 0

        min_red = self._search_min(self.red, min_ts=pr_item.min_ts, max_ts=pr_item.max_ts)
        min_amber = amber_spans[0].min_ts

        if min_red is None or ts < min_red:
            return ts - min_amber

        pre_amber = min_red - min_amber
        return pre_amber

    def get_red_amber_time(self, ts):
        """
        :type ts: long
        :rtype: long
        """
        pr_list = filter(lambda time_span: ts in time_span, self.prohibiting)

        if not len(pr_list):
            # current collor is green
            return 0

        pr_item = pr_list[0]  # type: TimeSpan

        # find amber intervals in prohibited item
        amber_spans = filter(lambda span: span.check_overlap(pr_item), self.amber)  # type: list[TimeSpan]

        # amber started before ts
        amber_spans = filter(lambda span: span.min_ts < ts, amber_spans)  # type: list[TimeSpan]
        if not len(amber_spans):
            return 0

        min_red = self._search_min(self.red, min_ts=pr_item.min_ts, max_ts=pr_item.max_ts)
        if min_red and min_red < amber_spans[-1].min_ts:
            return ts - amber_spans[-1].min_ts

        else:
            return 0

    @staticmethod
    def _search_max(items, min_ts=None, max_ts=None):
        """
        :type items: list[TimeSpan]
        :type min_ts: long | None
        :type max_ts: long | None
        :rtype: long | None
        """
        if len(items) == 0:
            return None

        if max_ts is None:
            return items[-1].max_ts

        if min_ts is None:
            min_ts = 0

        for index in range(len(items) - 1, -1, -1):
            item = items[index]
            if min_ts < item.max_ts < max_ts:
                return item.max_ts

        return None

    @staticmethod
    def _search_min(items, min_ts=None, max_ts=None):
        """
        :type items: list[TimeSpan]
        :type min_ts: long | None
        :type max_ts: long | None
        :rtype: long | None
        """
        if len(items) == 0:
            return None

        if min_ts is None:
            return items[0].min_ts

        if max_ts is None:
            # add 1 to include self last item in case of opened interval
            max_ts = items[-1].min_ts + 1

        for item in items:  # type: TimeSpan
            if min_ts < item.min_ts < max_ts:
                return item.min_ts

        return None

    def _get_prohibiting(self):
        """
         :rtype: list[TimeSpan]
         """
        prohibiting_items = []

        # if amber or red catch ts < green cts ts - add first item to prohibiting list
        min_time = None
        if len(self.amber) and len(self.red):
            min_time = min(self.amber[0].min_ts, self.red[0].min_ts)
        elif len(self.amber):
            min_time = self.amber[0].min_ts
        elif len(self.red):
            min_time = self.red[0].min_ts

        if min_time is not None and len(self.green) and min_time < self.green[0].min_ts:
            max_time_red = self._search_max(self.red, max_ts=self.green[0].min_ts)
            max_time_amber = self._search_max(self.amber, max_ts=self.green[0].min_ts)
            max_time = max(max_time_amber, max_time_red)
            if None not in [max_time, min_time]:
                prohibiting_items.append(TimeSpan(min_time, max_time))

        # all spaces between green are prohibiting intervals
        for index in range(0, len(self.green) - 1, 1):
            max_time_red = self._search_max(self.red, max_ts=self.green[index + 1].min_ts)
            max_time_amber = self._search_max(self.amber, max_ts=self.green[index + 1].min_ts)
            max_time = max(max_time_amber, max_time_red)

            min_time_red = self._search_min(self.red, min_ts=self.green[index].max_ts,
                                            max_ts=self.green[index + 1].min_ts)

            min_time_amber = self._search_min(self.amber, min_ts=self.green[index].max_ts,
                                              max_ts=self.green[index + 1].min_ts)

            min_time = min(min_time_red, min_time_amber) or min_time_red or min_time_amber
            if None not in [max_time, min_time]:
                prohibiting_items.append(TimeSpan(min_time, max_time))

        # if amber or red lost ts > green lost ts - add one more item to prohibiting list
        max_time = None
        if len(self.amber) and len(self.red):
            max_time = max(self.amber[-1].max_ts, self.red[-1].max_ts)
        elif len(self.amber):
            max_time = self.amber[-1].max_ts
        elif len(self.red):
            max_time = self.red[-1].max_ts

        # max_time = max(self.amber[-1].max_ts, self.red[-1].max_ts)
        if max_time is not None and len(self.green) and self.green[-1].max_ts < max_time:
            min_time_red = self._search_min(self.red, min_ts=self.green[-1].max_ts)
            min_time_amber = self._search_min(self.amber, min_ts=self.green[-1].max_ts)
            min_time = min(min_time_red, min_time_amber) or min_time_red or min_time_amber
            if None not in [max_time, min_time]:
                prohibiting_items.append(TimeSpan(min_time, max_time))

        return prohibiting_items

    def check(self, time_span):
        """
        :type time_span: TimeSpan
        :rtype: list[TimeSpan]
        """
        spans = filter(lambda x: x.check_overlap(time_span), self.prohibiting)
        return spans


class TlLog:
    def __init__(self, time, channels, red_only=False):
        """
        :type time: TimeSpan
        :type channels: list[_TrafficLight]
        :type red_only: bool
        """
        self.channels = channels
        self.time = time
        self.red_only = red_only

    @staticmethod
    def deserialize(tl_ident, jnode, red_only=False):
        """
        :type tl_ident: int
        :type jnode: dict
        :type red_only: bool
        :rtype: TlLog

        :raise JnodeError:
        """
        ts_jnode = jnode.get('ts')

        if type(ts_jnode) != list or len(ts_jnode) != 2:
            eventlog.log_e_param('#/ts')
            raise JnodeError

        time_interval = TimeSpan(*ts_jnode)

        traffic_lights_jnode = jnode.get('tl')
        if type(traffic_lights_jnode) != list:
            eventlog.log_e_param('#/tl')

        traffic_lights = []
        for tl_jnode in traffic_lights_jnode:
            if tl_ident != tl_jnode['id']:
                continue

            red = tl_jnode.get('red', [])

            amber = []
            green = []
            if not red_only:
                amber = tl_jnode.get('amber', [])
                green = tl_jnode.get('green', [])

            red_intervals = map(TimeSpan.create, red)
            amber_intervals = map(TimeSpan.create, amber)
            green_intervals = map(TimeSpan.create, green)

            traffic_light = _TrafficLight(tl_ident, red_intervals, amber_intervals, green_intervals, red_only)
            traffic_lights.append(traffic_light)

        if not traffic_lights:
            eventlog.log_d('fail: duplicate idents found')
            raise JnodeError

        return TlLog(time_interval, traffic_lights, red_only)


class _TlService(ITlService):
    """
    :type _tl_log: TlLog | None
    """
    def __init__(self, twisted_reactor, args):
        """
        :type twisted_reactor: twisted.internet.interfaces.IReactorUNIX
        :type args: JsonTlServiceArgs
        """
        ITlService.__init__(self)

        self._twisted_reactor = twisted_reactor

        self._tl_log = None
        self._tl_ident = args.tl_ident
        self._red_only = args.red_only

        self._endpoint = args.endpoint
        self._inotify = inotify.INotify()
        self._inotify.startReading()
        self._reset_inotify_watch()

    def destroy(self):
        try:
            self._inotify.ignore(FilePath(self._endpoint))
        except KeyError:
            pass

    def _reset_inotify_watch(self):
        try:
            self._inotify.ignore(FilePath(self._endpoint))
        except KeyError:
            pass

        fs.silent_mkdir_r(self._endpoint)

        try:
            self._inotify.watch(
                FilePath(self._endpoint),
                mask=inotify.IN_MOVED_TO | inotify.IN_CLOSE_WRITE | inotify.IN_DELETE_SELF | inotify.IN_MOVE_SELF,
                callbacks=[self._handle_react])
        except inotify.INotifyError as x:
            eventlog.log_d('fail: reset inotify watch (%s)' % x)

    # noinspection PyUnusedLocal
    def _handle_react(self, ignored, filepath, mask):
        if (mask & inotify.IN_DELETE_SELF) or (mask & inotify.IN_MOVE_SELF):
            self._twisted_reactor.callLater(0, self._reset_inotify_watch)
            eventlog.log_w('endpoint was changed to %s' % self._endpoint)
            return

        try:
            react_jnode = fs.load_json3(filepath.path, default=None)
            if react_jnode:
                self._tl_log = TlLog.deserialize(self._tl_ident, react_jnode, self._red_only)

        except JnodeError as x:
            return

        except Exception as x:
            eventlog.log_x()
            return

        finally:
            fs.silent_rm_r(filepath.path)

    def check(self, min_ts, max_ts):
        """
        :type min_ts: long
        :type max_ts: long
        :rtype: list[TimeSpan]
        """
        if self._tl_log is None:
            return []

        time_interval = TimeSpan(min_ts, max_ts)
        if not self._tl_log.time.check_overlap(time_interval):
            # there is no stats for [cts ,lts]
            return []

        ch_list = filter(lambda ch: ch.ch_id == self._tl_ident, self._tl_log.channels)

        if len(ch_list) != 1:
            eventlog.log_d("fail: find channel")
            return []

        ch_args = ch_list[0]  # type: _TrafficLight

        violation_list = ch_args.check(time_interval)
        return violation_list

    def get_times(self, ts):
        """
        :param ts: value in usec
        :type ts: long

        :return: (green-amber-time, red-time, red-amber-time). Values in usec
        :rtype: (long, long, long)
        """
        traffic_lights = filter(lambda ch: ch.ch_id == self._tl_ident, self._tl_log.channels)
        assert 1 == len(traffic_lights)

        tl = traffic_lights[0]  # type: _TrafficLight
        return tl.get_green_amber_time(ts), tl.get_red_time(ts), tl.get_red_amber_time(ts)
