from collections import defaultdict
from os.path import isabs


class SceneReactArgs:
    def __init__(self, mc_dir, mg_dir, lr_dir, ra_dir, tout_msec):
        """
        :type mc_dir: str
        :type mg_dir: str
        :type lr_dir: str
        :type ra_dir: str
        :type tout_msec: str
        """
        self.mc_dir = mc_dir
        self.mg_dir = mg_dir
        self.lr_dir = lr_dir
        self.ra_dir = ra_dir
        self.tout_msec = tout_msec

    @staticmethod
    def deserialize(scene_react_jnode):
        """
        :type scene_react_jnode: dict
        :rtype _SceneReactArgs
        """
        mc_dir = scene_react_jnode.get('McDir', '')
        mg_dir = scene_react_jnode.get('MgDir', '')
        lr_dir = scene_react_jnode.get('LrDir', '')
        ra_dir = scene_react_jnode.get('RaDir', '')
        tout = scene_react_jnode['Tout']

        if not mc_dir and not mg_dir:
            raise ValueError('At least one of mc_dir or mg_dir should be defined')

        if not lr_dir:
            raise ValueError('The lr_dir should be defined')

        args = SceneReactArgs(mc_dir, mg_dir, lr_dir, ra_dir, tout)
        return args


class TrackReactArgs:
    def __init__(self, react_dir):
        """
        :type react_dir: str
        """
        self.react_dir = react_dir

    @staticmethod
    def deserialize(track_react_jnode):
        """
        :type track_react_jnode: dict
        :rtype: TrackReactArgs
        """
        react_dir = track_react_jnode['Dir']
        args = TrackReactArgs(react_dir)
        return args


class SceneEventArgs:
    def __init__(self, event_dirs, skip_rate):
        """
        :type event_dirs: set of str
        :type skip_rate: int
        """
        self.event_dirs = event_dirs
        self.skip_rate = skip_rate

    @staticmethod
    def deserialize(event_dirs, skip_rate):
        """
        :type event_dirs: set of str
        :type skip_rate: int
        :rtype: SceneEventArgs
        """
        args = SceneEventArgs(event_dirs, skip_rate)
        return args


class ImageQueueArgs:
    def __init__(self, endpoint, min_file_count, max_file_count):
        """
        :type endpoint: str
        :type min_file_count: int
        :type max_file_count: int
        """
        self.endpoint = endpoint
        self.min_file_count = min_file_count
        self.max_file_count = max_file_count
        self.log_level = 'info'

    @staticmethod
    def deserialize(store_jnode):
        """
        :type store_jnode: dict
        :rtype: ImageQueueArgs
        """
        endpoint = store_jnode.get('endpoint')
        if not (unicode == type(endpoint) and len(endpoint) < 64 and isabs(endpoint)):
            raise ValueError('main.conf#/store/endpoint')

        max_file_count = store_jnode.get('max-file-count', 15)
        if not (int == type(max_file_count) and 15 <= max_file_count <= 50):
            raise ValueError('main.conf#/store/max-file-count')

        return ImageQueueArgs(endpoint, max_file_count - 5, max_file_count)


class TrackEventArgs:
    def __init__(self, debounce_tout, event_dirs):
        """
        :type debounce_tout
        :type event_dirs: list of str
        """
        self.debounce_tout = debounce_tout
        self.event_dirs = event_dirs

    @staticmethod
    def deserialize(track_event_jnode):
        """
        :type track_event_jnode: dict
        :rtype: TrackEventArgs
        """
        debounce_tout = track_event_jnode.get('debounce-tout', 3)
        endpoints = track_event_jnode.get('endpoints', [])

        if not (0 <= debounce_tout <= 30):
            raise ValueError('main.conf#/fact-event/debounce-tout')
        args = TrackEventArgs(debounce_tout=debounce_tout, event_dirs=endpoints)
        return args


class LrUpscaleArgs:
    def __init__(self, left, top, right, bottom):
        """
        :type left: float
        :type top: float
        :type right: float
        :type bottom: float
        """
        self.left = left
        self.top = top
        self.right = right
        self.bottom = bottom

        self._jnode = [left, top, right, bottom]

    def serialize(self):
        """
        :rtype: list of float
        """
        return self._jnode


class StatCountersArgs:
    def __init__(self, stat_filepath, stat_tout_sec):
        self.stat_filepath = stat_filepath
        self.stat_tout_sec = stat_tout_sec

    @staticmethod
    def deserialize(monitoring_jnode):
        """
        :type monitoring_jnode: dict
        :rtype: StatCountersArgs
        """
        stat_filepath = monitoring_jnode.get('StatFile')
        if not (unicode == type(stat_filepath) and len(stat_filepath) < 64 and isabs(stat_filepath)):
            raise ValueError('main.conf#/Monitoring/StatFile')

        raw_stat_tout_sec = monitoring_jnode.get('StatTout')
        if not (int == type(raw_stat_tout_sec) and 0 <= raw_stat_tout_sec <= 255):
            raise ValueError('main.conf#/Monitoring/StatTout')

        args = StatCountersArgs(stat_filepath, raw_stat_tout_sec)
        return args


class KernelArgs:
    DEF_LR_RA_OID_MATCH_STORE_TOUT_SEC = 5
    DEF_RA_TRACK_STORE_TOUT_SEC = 5
    DEF_MIN_RATRACK_LEN = 5

    def __init__(self, stat_counters_args, scene_react_args, track_react_args, scene_event_args_list, track_event_args,
                 lr_ra_oid_match_store_tout, ra_track_store_tout, min_ra_track_len, lr_upscale_args, image_store_args,
                 systest=False, plate_detected_mode=False):
        """
        :type stat_counters_args: StatCountersArgs
        :type scene_react_args: SceneReactArgs
        :type track_react_args: TrackReactArgs
        :type scene_event_args_list: list of SceneEventArgs
        :type track_event_args: TrackEventArgs
        :type lr_ra_oid_match_store_tout: int
        :type ra_track_store_tout: int
        :type min_ra_track_len: int
        :type image_store_args: ImageQueueArgs | None
        :type systest: bool
        :type plate_detected_mode: bool
        """
        self.stat_counters_args = stat_counters_args

        self.scene_react_args = scene_react_args
        self.track_react_args = track_react_args

        self.scene_event_args_list = scene_event_args_list
        self.track_event_args = track_event_args

        self.lr_ra_oid_match_store_tout = lr_ra_oid_match_store_tout
        self.ra_track_store_tout = ra_track_store_tout
        self.min_ra_track_len = min_ra_track_len

        self.lr_upscale_args = lr_upscale_args
        self.image_store_args = image_store_args
        self.systest = systest
        self.plate_detected_mode = plate_detected_mode

    @staticmethod
    def deserialize(main_conf_jnode, misc_conf_jnode):
        """
        :type main_conf_jnode: dict
        :type misc_conf_jnode: dict
        :rtype: KernelArgs
        """
        # define stat-counters-args
        monitoring_jnode = main_conf_jnode.get('Monitoring')
        if not monitoring_jnode:
            raise ValueError('main.conf#/Monitoring')

        stat_counters_args = StatCountersArgs.deserialize(monitoring_jnode)

        # define others
        file_api_jnode = main_conf_jnode['FileApi']
        track_event_args = TrackEventArgs.deserialize(file_api_jnode['TrackEvent'])
        scene_react_args = SceneReactArgs.deserialize(file_api_jnode['AddSceneReact'])
        track_react_args = TrackReactArgs.deserialize(file_api_jnode['AddTrackReact'])
        image_store_jnode = main_conf_jnode.get('store')
        image_store_args = ImageQueueArgs.deserialize(image_store_jnode) if image_store_jnode else None

        skip_rate_to_dir_set_map = defaultdict(set)
        for entry in file_api_jnode['SceneEvent']:
            skip_rate_to_dir_set_map[entry["SkipRate"]].add(entry["Dir"])

        scene_event_args_list = map(
            lambda (skip_rate, event_dirs): SceneEventArgs.deserialize(event_dirs, skip_rate),
            skip_rate_to_dir_set_map.items())

        lr_ra_oid_match_store_tout = misc_conf_jnode.get(
            'lpr_track_caching_time', KernelArgs.DEF_LR_RA_OID_MATCH_STORE_TOUT_SEC)

        ra_track_store_tout = misc_conf_jnode.get(
            'radar_track_caching_time', KernelArgs.DEF_RA_TRACK_STORE_TOUT_SEC)

        min_ra_track_len = misc_conf_jnode.get(
            'min_valid_radar_track_length', KernelArgs.DEF_MIN_RATRACK_LEN)

        lr_upscale_args = LrUpscaleArgs(left=2.5, top=18.0, right=2.5, bottom=5.0)

        systest = main_conf_jnode.get('systest', False)
        plate_detected_mode = main_conf_jnode.get('plate-detected-mode', False)

        kernel_args = KernelArgs(
            stat_counters_args, scene_react_args, track_react_args, scene_event_args_list, track_event_args,
            lr_ra_oid_match_store_tout, ra_track_store_tout, min_ra_track_len, lr_upscale_args, image_store_args,
            systest, plate_detected_mode)

        return kernel_args
