from _models import RaScene, RaTrack


class _RaTrackContext:
    def __init__(self, ra_track, rm_delayed_call):
        """
        :type ra_track: RaTrack
        :type rm_delayed_call: twisted.internet.interfaces.IDelayedCall
        """
        self.ra_track = ra_track
        self.rm_delayed_call = rm_delayed_call
        self.rm_request = False


class RaTrackStore:
    def __init__(self, twisted_reactor, tout_sec):
        """
        :type twisted_reactor: twisted.internet.interfaces.IReactorTime
        :type tout_sec: int
        """
        self._twisted_reactor = twisted_reactor
        self._tout_sec = tout_sec
        self._ra_oid_to_context_map = {}

        self._tout_handler = None

    def set_tout_handler(self, handler):
        self._tout_handler = handler

    def _handle_ra_track_tout(self, ra_oid):
        ra_track_context = self._ra_oid_to_context_map.get(ra_oid, None)
        if not ra_track_context.rm_request:
            if self._tout_handler:
                self._tout_handler(ra_track_context.ra_track)

        self._ra_oid_to_context_map.pop(ra_oid, None)

    def add_ra_scene(self, ra_scene):
        """
        :type ra_scene: RaScene
        """
        for ra_item in ra_scene.scene:
            ra_oid = ra_item.oid

            context = self._ra_oid_to_context_map.get(ra_oid, None)
            if context:
                # Update existing ra_track
                # if not ra_track_context.rm_delayed_call.active():
                #    pass
                context.rm_delayed_call.reset(self._tout_sec)
                context.ra_track.append(ra_item)

            else:
                # Create new ra_track.
                ra_track = RaTrack(ra_item)

                # Schedule to remove ra_track after _ra_track_tout_sec since the last update.
                rm_delayed_call = self._twisted_reactor.callLater(
                    self._tout_sec, self._handle_ra_track_tout, ra_oid)

                self._ra_oid_to_context_map[ra_oid] = _RaTrackContext(ra_track, rm_delayed_call)

    def rm_ra_track(self, ra_oid):
        """
        :type ra_oid: int
        """
        ra_track_context = self._ra_oid_to_context_map.get(ra_oid, None)
        if ra_track_context:
            # ra_track will be removed automaticaly after _ra_track_tout_sec timeout since the last update.
            # rm_request flag indicates that it's necessary to hide ra_track presence in the store for its clients.
            ra_track_context.rm_request = True

    def try_get_ra_track(self, ra_oid):
        """
        :type ra_oid: int
        :rtype: RaTrack | None
        """
        ra_track_context = self._ra_oid_to_context_map.get(ra_oid, None)
        if ra_track_context and not ra_track_context.rm_request:
            return ra_track_context.ra_track

        return None
